import os
import sys
from hub.constants import KB, MB
from hub.util.exceptions import (
    SampleCompressionError,
    TensorMetaMissingRequiredValue,
    TensorMetaMutuallyExclusiveKeysError,
    UnsupportedCompressionError,
    SampleHtypeMismatchError,
)
import pytest
from hub.core.tensor import Tensor
from hub.tests.common import TENSOR_KEY, assert_images_close
import numpy as np

import hub
from hub.core.dataset import Dataset


def _populate_compressed_samples(tensor: Tensor, cat_path, flower_path, count=1):
    for _ in range(count):
        tensor.append(hub.read(cat_path))
        tensor.append(hub.read(flower_path))
        tensor.append(np.ones((100, 100, 4), dtype="uint8"))
        tensor.append(
            np.ones((100, 100, 4), dtype=int).tolist()
        )  # test safe downcasting of python scalars

        tensor.extend(
            [
                hub.read(flower_path),
                hub.read(cat_path),
            ]
        )


def test_populate_compressed_samples(local_ds, cat_path, flower_path):
    images = local_ds.create_tensor(
        TENSOR_KEY,
        htype="image",
        sample_compression="png",
        max_chunk_size=2 * MB,
        tiling_threshold=1 * MB,
    )

    assert images.meta.dtype == "uint8"
    assert images.meta.sample_compression == "png"

    _populate_compressed_samples(images, cat_path, flower_path)

    expected_shapes = [
        (900, 900, 3),
        (513, 464, 4),
        (100, 100, 4),
        (100, 100, 4),
        (513, 464, 4),
        (900, 900, 3),
    ]

    assert len(images) == 6

    for img, exp_shape in zip(images, expected_shapes):
        arr = img.numpy()
        assert arr.shape == exp_shape
        assert arr.dtype == "uint8"

    assert images.shape == (6, None, None, None)
    assert images.shape_interval.lower == (6, 100, 100, 3)
    assert images.shape_interval.upper == (6, 900, 900, 4)


def test_iterate_compressed_samples(local_ds, cat_path, flower_path):
    images = local_ds.create_tensor(TENSOR_KEY, htype="image", sample_compression="png")

    assert images.meta.dtype == "uint8"
    assert images.meta.sample_compression == "png"

    _populate_compressed_samples(images, cat_path, flower_path)

    expected_shapes = [
        (900, 900, 3),
        (513, 464, 4),
        (100, 100, 4),
        (100, 100, 4),
        (513, 464, 4),
        (900, 900, 3),
    ]

    assert len(images) == len(expected_shapes)
    for image, expected_shape in zip(images, expected_shapes):
        x = image.numpy()

        assert (
            type(x) == np.ndarray
        ), "Check is necessary in case a `PIL` object is returned instead of an array."
        assert x.shape == expected_shape
        assert x.dtype == "uint8"


def test_uncompressed(local_ds):
    images = local_ds.create_tensor(TENSOR_KEY, sample_compression=None)

    images.append(np.ones((100, 100, 100)))
    images.extend(np.ones((3, 101, 2, 1)))
    local_ds.clear_cache()
    np.testing.assert_array_equal(images[0].numpy(), np.ones((100, 100, 100)))
    np.testing.assert_array_equal(images[1:4].numpy(), np.ones((3, 101, 2, 1)))


def test_byte_sample_compression(memory_ds):
    with memory_ds as ds:
        ds.create_tensor("xyz", sample_compression="lz4")
        for i in range(10):
            ds.xyz.append(i * np.ones((100, 100, 100)))

    for i in range(10):
        np.testing.assert_array_equal(ds.xyz[i].numpy(), i * np.ones((100, 100, 100)))


@pytest.mark.xfail(raises=SampleCompressionError, strict=True)
@pytest.mark.parametrize(
    "bad_shape",
    [
        # raises OSError: cannot write mode LA as JPEG
        (100, 100, 2),
        # raises OSError: cannot write mode RGBA as JPE
        (100, 100, 4),
    ],
)
def test_jpeg_bad_shapes(memory_ds: Dataset, bad_shape):
    # jpeg allowed shapes:
    # ---------------------
    # (100) works!
    # (100,) works!
    # (100, 100) works!
    # (100, 100, 1) works!
    # (100, 100, 2) raises   | OSError: cannot write mode LA as JPEG
    # (100, 100, 3) works!
    # (100, 100, 4) raises   | OSError: cannot write mode RGBA as JPEG
    # (100, 100, 5) raises   | TypeError: Cannot handle this data type: (1, 1, 5), |u1
    # (100, 100, 100) raises | TypeError: Cannot handle this data type: (1, 1, 100), |u1

    tensor = memory_ds.create_tensor(TENSOR_KEY, sample_compression="jpeg")
    tensor.append(np.ones(bad_shape, dtype="uint8"))


def test_compression_aliases(memory_ds: Dataset):
    tensor = memory_ds.create_tensor("jpeg_tensor", sample_compression="jpeg")
    assert tensor.meta.sample_compression == "jpeg"

    tensor = memory_ds.create_tensor("jpg_tensor", sample_compression="jpg")
    assert tensor.meta.sample_compression == "jpeg"

    tensor = memory_ds.create_tensor("tiff_tensor", sample_compression="tiff")
    assert tensor.meta.sample_compression == "tiff"

    tensor = memory_ds.create_tensor("tif_tensor", sample_compression="tif")
    assert tensor.meta.sample_compression == "tiff"


@pytest.mark.xfail(raises=UnsupportedCompressionError, strict=True)
def test_unsupported_compression(memory_ds: Dataset):
    memory_ds.create_tensor(TENSOR_KEY, sample_compression="bad_compression")
    # TODO: same tests but with `dtype`


@pytest.mark.xfail(raises=TensorMetaMissingRequiredValue, strict=True)
def test_missing_sample_compression_for_image(memory_ds: Dataset):
    memory_ds.create_tensor("tensor", htype="image")


@pytest.mark.xfail(raises=TensorMetaMutuallyExclusiveKeysError, strict=True)
def test_sample_chunk_compression_mutually_exclusive(memory_ds: Dataset):
    memory_ds.create_tensor(
        "tensor", htype="image", sample_compression="png", chunk_compression="lz4"
    )


def test_chunkwise_compression(memory_ds, cat_path, flower_path):
    ds = memory_ds
    im_ct = 5
    chunk_size = 600 * KB
    with ds:
        images = ds.create_tensor(
            "images",
            htype="image",
            chunk_compression="jpg",
            max_chunk_size=chunk_size,
            tiling_threshold=chunk_size,
        )
        images.extend([hub.read(cat_path)] * im_ct)
        expected_arr = np.random.randint(0, 10, (500, 450, 3)).astype("uint8")
        images.append(expected_arr)
        images.extend([hub.read(cat_path)] * im_ct)
        expected_img = np.array(hub.read(cat_path))
    ds.clear_cache()
    for i, img in enumerate(images):
        if i == im_ct:
            assert_images_close(img.numpy(), expected_arr)
        else:
            assert_images_close(img.numpy(), expected_img)
    with ds:
        images = ds.create_tensor(
            "images2",
            htype="image",
            chunk_compression="png",
            max_chunk_size=chunk_size,
            tiling_threshold=chunk_size,
        )
        images.extend([hub.read(flower_path)] * im_ct)
        expected_arr = np.random.randint(0, 256, (200, 250, 4)).astype("uint8")
        images.append(expected_arr)
        images.extend([hub.read(flower_path)] * im_ct)
        expected_img = np.array(hub.read(flower_path))
    ds.clear_cache()
    for i, img in enumerate(images):
        if i == im_ct:
            assert_images_close(img.numpy(), expected_arr)
        else:
            assert_images_close(img.numpy(), expected_img)
    with ds:
        labels = ds.create_tensor(
            "labels",
            chunk_compression="lz4",
            max_chunk_size=chunk_size,
            tiling_threshold=chunk_size,
        )
        data = [
            np.random.randint(0, 256, (150, 150)).astype("uint8") for _ in range(20)
        ]
        labels.extend(data)
    ds.clear_cache()
    for row, label in zip(data, labels):
        np.testing.assert_array_equal(row, label.numpy())

    data = np.random.randint(0, 256, (5, 1500, 1500)).astype("uint8")
    with ds:
        ds.labels.extend(data)  # type: ignore
    ds.clear_cache()
    assert len(ds.labels) == 25
    for i in range(5):
        np.testing.assert_array_equal(data[i], ds.labels[20 + i].numpy())


def _decompress_audio_helper(path):
    import av  # type: ignore

    container = av.open(path)
    for frame in container.decode(audio=0):
        if not frame.is_corrupt:
            audio = frame.to_ndarray().astype("<f4")
            break

    for frame in container.decode(audio=0):
        audio = np.concatenate((audio, frame.to_ndarray().astype("<f4")), axis=1)

    return np.transpose(audio)


@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
@pytest.mark.parametrize("compression", hub.compression.AUDIO_COMPRESSIONS)
def test_audio(local_ds, compression, audio_paths):
    path = audio_paths[compression]
    arr = _decompress_audio_helper(path)
    local_ds.create_tensor("audio", htype="audio", sample_compression=compression)
    with local_ds:
        for _ in range(10):
            local_ds.audio.append(hub.read(path))  # type: ignore
    for i in range(10):
        decompressed = local_ds.audio[i].numpy()
        np.testing.assert_array_equal(decompressed[: len(arr), :], arr)  # type: ignore


def test_exif(memory_ds, compressed_image_paths):
    ds = memory_ds
    with ds:
        ds.create_tensor("images", htype="image", sample_compression="jpeg")
        for path in compressed_image_paths["jpeg"]:
            ds.images.append(hub.read(path))
    for image in ds.images:
        assert isinstance(image.sample_info["exif"], dict), (
            type(image.sample_info["exif"]),
            path,
        )


def test_forced_htypes(
    memory_ds, grayscale_image_paths, color_image_paths, flower_path
):
    with memory_ds as ds:
        gray = ds.create_tensor("gray", htype="image.gray", sample_compression="jpeg")
        rgb = ds.create_tensor("rgb", htype="image.rgb", sample_compression="jpeg")

        gray.append(hub.read(grayscale_image_paths["jpeg"]))
        gray.append(hub.read(color_image_paths["jpeg"]))
        gray.append(hub.read(flower_path))
        gray.extend(np.ones((4, 10, 10, 3), dtype=np.uint8))
        gray.extend(
            [hub.read(color_image_paths["jpeg"]), np.ones((10, 10), dtype=np.uint8)]
        )

        rgb.append(hub.read(grayscale_image_paths["jpeg"]))
        rgb.append(hub.read(color_image_paths["jpeg"]))
        rgb.append(hub.read(flower_path))
        rgb.extend(np.ones((4, 10, 10), dtype=np.uint8))
        rgb.extend(
            [
                hub.read(grayscale_image_paths["jpeg"]),
                np.ones((10, 10, 3), dtype=np.uint8),
            ]
        )

        gray_png = ds.create_tensor(
            "gray_png", htype="image.gray", sample_compression="png"
        )
        rgb_png = ds.create_tensor(
            "rgb_png", htype="image.rgb", sample_compression="png"
        )

        gray_png.append(hub.read(flower_path))
        gray_png.append(np.ones((10, 10, 4), dtype=np.uint8))

        rgb_png.append(hub.read(flower_path))
        rgb_png.append(np.ones((10, 10, 4), dtype=np.uint8))

        with pytest.raises(SampleHtypeMismatchError):
            rgb_png.append(1)

        with pytest.raises(TensorMetaMissingRequiredValue):
            ds.create_tensor("abc", htype="image.rgb")

        with pytest.raises(TensorMetaMissingRequiredValue):
            ds.create_tensor("abc", htype="image.gray")

    for sample in gray:
        assert len(sample.shape) == 2

    for sample in rgb:
        assert len(sample.shape) == 3

    for sample in gray_png:
        assert len(sample.shape) == 2

    for sample in rgb_png:
        assert len(sample.shape) == 3



import pathlib
from typing import Optional, Union
from hub.core.storage.provider import StorageProvider
from hub.util.tag import process_hub_path
from hub.constants import HUB_CLOUD_DEV_USERNAME
import glob
import os


def is_hub_cloud_path(path: str):
    """Whether given ``path`` is a hub cloud path."""
    return path.startswith("hub://")


def get_path_from_storage(storage) -> str:
    """Extracts the underlying path from a given storage."""
    from hub.core.storage.lru_cache import LRUCache

    if isinstance(storage, LRUCache):
        return get_path_from_storage(storage.next_storage)
    elif isinstance(storage, StorageProvider):
        if hasattr(storage, "hub_path"):
            return storage.hub_path  # type: ignore
        return storage.root
    else:
        raise ValueError("Invalid storage type.")


def find_root(path: str) -> str:
    """Find the root of the dataset within the given path.

    Note:
        The "root" is defined as the subdirectory (within path) that has > 1 folder/file (if applicable).
        in other words, if there is a directory structure like:
        dataset -
            Images -
                class1 -
                    img.jpg
                    ...
                class2 -
                    img.jpg
                    ...
                ...

        root is "dataset/Images"

    Args:
        path (str): The local path to folder containing an unstructured dataset and of the form ./path/to/dataset or ~/path/to/dataset or path/to/dataset.

    Returns:
        str: Root path of the unstructured dataset.
    """

    subs = glob.glob(os.path.join(path, "*"))

    subs = [
        sub for sub in subs if os.path.isdir(sub)
    ]  # only keep directories (ignore files)

    if len(subs) == 1:
        return find_root(subs[0])

    return path


def get_path_type(path: Optional[str]) -> str:
    if not isinstance(path, str):
        path = str(path)
    if path.startswith("http://") or path.startswith("https://"):
        return "http"
    elif path.startswith("gcs://") or path.startswith("gcp://"):
        return "gcs"
    elif path.startswith("s3://"):
        return "s3"
    elif path.startswith("gdrive://"):
        return "gdrive"
    else:
        return "local"


def is_remote_path(path: str) -> bool:
    return get_path_type(path) != "local"


def convert_string_to_pathlib_if_needed(path, convert_to_pathlib=False):
    converted_path = pathlib.Path(path)
    if convert_to_pathlib and "//" not in path:
        return converted_path
    return path


def convert_pathlib_to_string_if_needed(path: Union[str, pathlib.Path]) -> str:
    if isinstance(path, pathlib.Path):
        path = str(path)
    return path


def get_org_id_and_ds_name(path):
    if is_hub_cloud_path(path):
        _, org_id, ds_name, subdir = process_hub_path(path)
        if subdir:
            ds_name += "/" + subdir
    else:
        org_id = HUB_CLOUD_DEV_USERNAME
        ds_name = path.replace("/", "_").replace(".", "")

    return org_id, ds_name



from uuid import uuid4
import hub
from typing import Callable, List, Optional
from itertools import repeat
from hub.core.compute.provider import ComputeProvider
from hub.core.storage.memory import MemoryProvider
from hub.util.bugout_reporter import hub_reporter
from hub.util.compute import get_compute_provider
from hub.util.dataset import try_flushing
from hub.util.remove_cache import get_base_storage, get_dataset_with_zero_size_cache
from hub.util.transform import (
    check_lengths,
    check_transform_data_in,
    check_transform_ds_out,
    create_slices,
    delete_overwritten_chunks,
    get_lengths_generated,
    get_old_chunk_paths,
    get_pbar_description,
    process_transform_result,
    sanitize_workers_scheduler,
    store_data_slice,
    store_data_slice_with_pbar,
)
from hub.util.encoder import merge_all_meta_info
from hub.util.exceptions import (
    HubComposeEmptyListError,
    HubComposeIncompatibleFunction,
    TransformError,
)
from hub.hooks import dataset_written, dataset_read
from hub.util.version_control import auto_checkout, load_meta
from hub.util.class_label import sync_labels
import numpy as np


class ComputeFunction:
    def __init__(self, func, args, kwargs, name: Optional[str] = None):
        """Creates a ComputeFunction object that can be evaluated using .eval or used as a part of a Pipeline."""
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.name = self.func.__name__ if name is None else name

    def eval(
        self,
        data_in,
        ds_out: Optional[hub.Dataset] = None,
        num_workers: int = 0,
        scheduler: str = "threaded",
        progressbar: bool = True,
        skip_ok: bool = False,
        check_lengths: bool = True,
        pad_data_in: bool = False,
    ):
        """Evaluates the ComputeFunction on data_in to produce an output dataset ds_out.

        Args:
            data_in: Input passed to the transform to generate output dataset. Should support \__getitem__ and \__len__. Can be a Hub dataset.
            ds_out (Dataset, optional): The dataset object to which the transform will get written. If this is not provided, data_in will be overwritten if it is a Hub dataset, otherwise error will be raised.
                It should have all keys being generated in output already present as tensors. It's initial state should be either:-
                - Empty i.e. all tensors have no samples. In this case all samples are added to the dataset.
                - All tensors are populated and have sampe length. In this case new samples are appended to the dataset.
            num_workers (int): The number of workers to use for performing the transform. Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.
            scheduler (str): The scheduler to be used to compute the transformation. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.
                Defaults to 'threaded'.
            progressbar (bool): Displays a progress bar if True (default).
            skip_ok (bool): If True, skips the check for output tensors generated. This allows the user to skip certain tensors in the function definition.
                This is especially useful for inplace transformations in which certain tensors are not modified. Defaults to False.
            check_lengths (bool): If True, checks whether ds_out has tensors of same lengths initially.
            pad_data_in (bool): NOTE: This is only applicable if data_in is a Hub dataset. If True, pads tensors of data_in to match the length of the largest tensor in data_in.
                Defaults to False.

        Raises:
            InvalidInputDataError: If data_in passed to transform is invalid. It should support \__getitem__ and \__len__ operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.
            InvalidOutputDatasetError: If all the tensors of ds_out passed to transform don't have the same length. Using scheduler other than "threaded" with hub dataset having base storage as memory as ds_out will also raise this.
            TensorMismatchError: If one or more of the outputs generated during transform contain different tensors than the ones present in 'ds_out' provided to transform.
            UnsupportedSchedulerError: If the scheduler passed is not recognized. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.
        """

        pipeline = Pipeline([self])
        pipeline.eval(
            data_in,
            ds_out,
            num_workers,
            scheduler,
            progressbar,
            skip_ok,
            check_lengths,
            pad_data_in,
        )

    def __call__(self, sample_in):
        return self.func(sample_in, *self.args, **self.kwargs)


class Pipeline:
    def __init__(self, functions: List[ComputeFunction]):
        """Takes a list of functions decorated using hub.compute and creates a pipeline that can be evaluated using .eval"""
        self.functions = functions

    def __len__(self):
        return len(self.functions)

    def eval(
        self,
        data_in,
        ds_out: Optional[hub.Dataset] = None,
        num_workers: int = 0,
        scheduler: str = "threaded",
        progressbar: bool = True,
        skip_ok: bool = False,
        check_lengths: bool = True,
        pad_data_in: bool = False,
    ):
        """Evaluates the pipeline on data_in to produce an output dataset ds_out.

        Args:
            data_in: Input passed to the transform to generate output dataset. Should support \__getitem__ and \__len__. Can be a Hub dataset.
            ds_out (Dataset, optional): The dataset object to which the transform will get written. If this is not provided, data_in will be overwritten if it is a Hub dataset, otherwise error will be raised.
                It should have all keys being generated in output already present as tensors. It's initial state should be either:-
                - Empty i.e. all tensors have no samples. In this case all samples are added to the dataset.
                - All tensors are populated and have sampe length. In this case new samples are appended to the dataset.
            num_workers (int): The number of workers to use for performing the transform. Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.
            scheduler (str): The scheduler to be used to compute the transformation. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.
                Defaults to 'threaded'.
            progressbar (bool): Displays a progress bar if True (default).
            skip_ok (bool): If True, skips the check for output tensors generated. This allows the user to skip certain tensors in the function definition.
                This is especially useful for inplace transformations in which certain tensors are not modified. Defaults to False.
            check_lengths (bool): If True, checks whether ds_out has tensors of same lengths initially.
            pad_data_in (bool): NOTE: This is only applicable if data_in is a Hub dataset. If True, pads tensors of data_in to match the length of the largest tensor in data_in.
                Defaults to False.

        Raises:
            InvalidInputDataError: If data_in passed to transform is invalid. It should support \__getitem__ and \__len__ operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.
            InvalidOutputDatasetError: If all the tensors of ds_out passed to transform don't have the same length. Using scheduler other than "threaded" with hub dataset having base storage as memory as ds_out will also raise this.
            TensorMismatchError: If one or more of the outputs generated during transform contain different tensors than the ones present in 'ds_out' provided to transform.
            UnsupportedSchedulerError: If the scheduler passed is not recognized. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.
            TransformError: All other exceptions raised if there are problems while running the pipeline.
        """
        num_workers, scheduler = sanitize_workers_scheduler(num_workers, scheduler)
        overwrite = ds_out is None
        hub_reporter.feature_report(
            feature_name="eval",
            parameters={"Num_Workers": str(num_workers), "Scheduler": scheduler},
        )
        check_transform_data_in(data_in, scheduler)

        if isinstance(data_in, hub.Dataset):
            try_flushing(data_in)
            if overwrite:
                auto_checkout(data_in)
            original_data_in = data_in
            data_in = get_dataset_with_zero_size_cache(data_in)
            if pad_data_in:
                initial_padding_state = data_in._pad_tensors
                data_in._enable_padding()

        target_ds = data_in if overwrite else ds_out

        check_transform_ds_out(target_ds, scheduler, check_lengths)

        # if overwrite then we've already flushed and autocheckecked out data_in which is target_ds now
        if not overwrite:
            target_ds.flush()
            auto_checkout(target_ds)

        compute_provider = get_compute_provider(scheduler, num_workers)
        compute_id = str(uuid4().hex)
        target_ds._send_compute_progress(compute_id=compute_id, start=True, progress=0)

        initial_autoflush = target_ds.storage.autoflush
        target_ds.storage.autoflush = False
        progress_end_args = {"compute_id": compute_id, "progress": 100, "end": True}

        if not check_lengths:
            skip_ok = True

        try:
            self.run(
                data_in,
                target_ds,
                compute_provider,
                num_workers,
                scheduler,
                progressbar,
                overwrite,
                skip_ok,
            )
            target_ds._send_compute_progress(**progress_end_args, status="success")
        except Exception as e:
            target_ds._send_compute_progress(**progress_end_args, status="failed")
            raise TransformError(e) from e
        finally:
            compute_provider.close()
            if overwrite:
                original_data_in.storage.clear_cache_without_flush()
                load_meta(original_data_in)
                if pad_data_in and not initial_padding_state:
                    original_data_in._disable_padding()
            else:
                load_meta(target_ds)
                target_ds.storage.autoflush = initial_autoflush

    def run(
        self,
        data_in,
        target_ds: hub.Dataset,
        compute: ComputeProvider,
        num_workers: int,
        scheduler: str,
        progressbar: bool = True,
        overwrite: bool = False,
        skip_ok: bool = False,
    ):
        """Runs the pipeline on the input data to produce output samples and stores in the dataset.
        This receives arguments processed and sanitized by the Pipeline.eval method.
        """
        if isinstance(data_in, hub.Dataset):
            dataset_read(data_in)
        slices = create_slices(data_in, num_workers)
        storage = get_base_storage(target_ds.storage)
        class_label_tensors = [
            tensor.key
            for tensor in target_ds.tensors.values()
            if tensor.base_htype == "class_label"
            and not tensor.meta._disable_temp_transform
        ]
        label_temp_tensors = {}
        actual_tensors = (
            None
            if not class_label_tensors
            else [target_ds[t].key for t in target_ds.tensors]
        )

        for tensor in class_label_tensors:
            temp_tensor = f"_{tensor}_{uuid4().hex[:4]}"
            with target_ds:
                temp_tensor_obj = target_ds.create_tensor(
                    temp_tensor,
                    htype="class_label",
                    create_sample_info_tensor=False,
                    create_shape_tensor=False,
                    create_id_tensor=False,
                )
                temp_tensor_obj.meta._disable_temp_transform = True
                label_temp_tensors[tensor] = temp_tensor
            target_ds.flush()

        visible_tensors = list(target_ds.tensors)
        visible_tensors = [target_ds[t].key for t in visible_tensors]
        visible_tensors = list(set(visible_tensors) - set(class_label_tensors))

        tensors = list(target_ds._tensors())
        tensors = [target_ds[t].key for t in tensors]
        tensors = list(set(tensors) - set(class_label_tensors))

        group_index = target_ds.group_index
        version_state = target_ds.version_state
        if isinstance(storage, MemoryProvider):
            storages = [storage] * len(slices)
        else:
            storages = [storage.copy() for _ in slices]
        args = (
            group_index,
            tensors,
            visible_tensors,
            label_temp_tensors,
            actual_tensors,
            self,
            version_state,
            target_ds.link_creds,
            skip_ok,
        )
        map_inp = zip(slices, storages, repeat(args))

        if progressbar:
            desc = get_pbar_description(self.functions)
            result = compute.map_with_progressbar(
                store_data_slice_with_pbar,
                map_inp,
                total_length=len(data_in),
                desc=desc,
            )
        else:
            result = compute.map(store_data_slice, map_inp)
        result = process_transform_result(result)

        all_num_samples, all_tensors_generated_length = get_lengths_generated(
            result["tensor_metas"], tensors
        )

        check_lengths(all_tensors_generated_length, skip_ok)

        generated_tensors = [
            tensor for tensor, l in all_tensors_generated_length.items() if l > 0
        ]

        old_chunk_paths = get_old_chunk_paths(target_ds, generated_tensors, overwrite)
        merge_all_meta_info(
            target_ds, storage, generated_tensors, overwrite, all_num_samples, result
        )
        delete_overwritten_chunks(old_chunk_paths, storage, overwrite)
        dataset_written(target_ds)

        if label_temp_tensors:
            sync_labels(
                target_ds,
                label_temp_tensors,
                result["hash_label_maps"],
                num_workers=num_workers,
                scheduler=scheduler,
                verbose=progressbar,
            )


def compose(functions: List[ComputeFunction]):  # noqa: DAR101, DAR102, DAR201, DAR401
    """Takes a list of functions decorated using :func:`hub.compute` and creates a pipeline that can be evaluated using .eval

    Example::

        pipeline = hub.compose([my_fn(a=3), another_function(b=2)])
        pipeline.eval(data_in, ds_out, scheduler="processed", num_workers=2)

    The ``eval`` method evaluates the pipeline/transform function.

    It has the following arguments:

    - ``data_in``: Input passed to the transform to generate output dataset.

        - It should support ``__getitem__`` and ``__len__``. This can be a Hub dataset.

    - ``ds_out (Dataset, optional)``: The dataset object to which the transform will get written.

        - If this is not provided, data_in will be overwritten if it is a Hub dataset, otherwise error will be raised.
        - It should have all keys being generated in output already present as tensors.
        - It's initial state should be either:

            - Empty i.e. all tensors have no samples. In this case all samples are added to the dataset.
            - All tensors are populated and have same length. In this case new samples are appended to the dataset.

    - ``num_workers (int)``: The number of workers to use for performing the transform.

        - Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.

    - ``scheduler (str)``: The scheduler to be used to compute the transformation.

        - Supported values include: 'serial', 'threaded', 'processed' and 'ray'. Defaults to 'threaded'.

    - ``progressbar (bool)``: Displays a progress bar if True (default).

    - ``skip_ok (bool)``: If True, skips the check for output tensors generated.

        - This allows the user to skip certain tensors in the function definition.
        - This is especially useful for inplace transformations in which certain tensors are not modified. Defaults to ``False``.

    It raises the following errors:

    - ``InvalidInputDataError``: If data_in passed to transform is invalid. It should support ``__getitem__`` and ``__len__`` operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.

    - ``InvalidOutputDatasetError``: If all the tensors of ds_out passed to transform don't have the same length. Using scheduler other than "threaded" with hub dataset having base storage as memory as ds_out will also raise this.

    - ``TensorMismatchError``: If one or more of the outputs generated during transform contain different tensors than the ones present in 'ds_out' provided to transform.

    - ``UnsupportedSchedulerError``: If the scheduler passed is not recognized. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.

    - ``TransformError``: All other exceptions raised if there are problems while running the pipeline.
    """
    if not functions:
        raise HubComposeEmptyListError
    for index, fn in enumerate(functions):
        if not isinstance(fn, ComputeFunction):
            raise HubComposeIncompatibleFunction(index)
    return Pipeline(functions)


def compute(
    fn,
    name: Optional[str] = None,
) -> Callable[..., ComputeFunction]:  # noqa: DAR101, DAR102, DAR201, DAR401
    """Compute is a decorator for functions.

    The functions should have atleast 2 argument, the first two will correspond to ``sample_in`` and ``samples_out``.

    There can be as many other arguments as required.

    The output should be appended/extended to the second argument in a hub like syntax.

    Any value returned by the fn will be ignored.

    Example::

        @hub.compute
        def my_fn(sample_in: Any, samples_out, my_arg0, my_arg1=0):
            samples_out.my_tensor.append(my_arg0 * my_arg1)

        # This transform can be used using the eval method in one of these 2 ways:-

        # Directly evaluating the method
        # here arg0 and arg1 correspond to the 3rd and 4th argument in my_fn
        my_fn(arg0, arg1).eval(data_in, ds_out, scheduler="threaded", num_workers=5)

        # As a part of a Transform pipeline containing other functions
        pipeline = hub.compose([my_fn(a, b), another_function(x=2)])
        pipeline.eval(data_in, ds_out, scheduler="processed", num_workers=2)

    The ``eval`` method evaluates the pipeline/transform function.

    It has the following arguments:

    - ``data_in``: Input passed to the transform to generate output dataset.

        - It should support ``__getitem__`` and ``__len__``. This can be a Hub dataset.

    - ``ds_out (Dataset, optional)``: The dataset object to which the transform will get written.

        - If this is not provided, data_in will be overwritten if it is a Hub dataset, otherwise error will be raised.
        - It should have all keys being generated in output already present as tensors.
        - It's initial state should be either:

            - Empty i.e. all tensors have no samples. In this case all samples are added to the dataset.
            - All tensors are populated and have same length. In this case new samples are appended to the dataset.

    - ``num_workers (int)``: The number of workers to use for performing the transform.

        - Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.

    - ``scheduler (str)``: The scheduler to be used to compute the transformation.

        - Supported values include: 'serial', 'threaded', 'processed' and 'ray'. Defaults to 'threaded'.

    - ``progressbar (bool)``: Displays a progress bar if True (default).

    - ``skip_ok (bool)``: If True, skips the check for output tensors generated.

        - This allows the user to skip certain tensors in the function definition.
        - This is especially useful for inplace transformations in which certain tensors are not modified. Defaults to ``False``.

    It raises the following errors:

    - ``InvalidInputDataError``: If data_in passed to transform is invalid. It should support ``__getitem__`` and ``__len__`` operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.

    - ``InvalidOutputDatasetError``: If all the tensors of ds_out passed to transform don't have the same length. Using scheduler other than "threaded" with hub dataset having base storage as memory as ds_out will also raise this.

    - ``TensorMismatchError``: If one or more of the outputs generated during transform contain different tensors than the ones present in 'ds_out' provided to transform.

    - ``UnsupportedSchedulerError``: If the scheduler passed is not recognized. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.

    - ``TransformError``: All other exceptions raised if there are problems while running the pipeline.
    """

    def inner(*args, **kwargs):
        return ComputeFunction(fn, args, kwargs, name)

    return inner



import warnings
from hub.core.compute.provider import ComputeProvider
from pathos.pools import ProcessPool  # type: ignore
from pathos.helpers import mp as pathos_multiprocess  # type: ignore


class ProcessProvider(ComputeProvider):
    def __init__(self, workers):
        self.workers = workers
        self.pool = ProcessPool(nodes=workers)
        self.manager = pathos_multiprocess.Manager()
        self._closed = False

    def map(self, func, iterable):
        return self.pool.map(func, iterable)

    def create_queue(self):
        return self.manager.Queue()

    def close(self):
        self.pool.close()
        self.pool.join()
        self.pool.clear()
        self._closed = True

    def __del__(self):
        if not self._closed:
            self.close()
            warnings.warn(
                "process pool thread leak. check compute provider is closed after use"
            )



import numpy as np
import os
import sys
import pickle
import hub
import pytest
from hub.client.client import HubBackendClient
from hub.constants import GCS_OPT, S3_OPT
from hub.core.link_creds import LinkCreds
from hub.core.meta.encode.creds import CredsEncoder
from hub.core.storage.gcs import GCSProvider
from hub.core.storage.s3 import S3Provider
from hub.tests.common import is_opt_true
from hub.util.exceptions import (
    ManagedCredentialsNotFoundError,
    TensorMetaInvalidHtype,
    UnableToReadFromUrlError,
)

from hub.util.htype import parse_complex_htype  # type: ignore


def test_complex_htype_parsing():
    with pytest.raises(ValueError):
        is_sequence, is_link, htype = parse_complex_htype("link")

    is_sequence, is_link, htype = parse_complex_htype("sequence")
    assert is_sequence
    assert not is_link
    assert htype == "generic"

    with pytest.raises(ValueError):
        is_sequence, is_link, htype = parse_complex_htype("sequence[link]")

    with pytest.raises(ValueError):
        is_sequence, is_link, htype = parse_complex_htype("link[sequence]")

    is_sequence, is_link, htype = parse_complex_htype("sequence[image]")
    assert is_sequence
    assert not is_link
    assert htype == "image"

    is_sequence, is_link, htype = parse_complex_htype("link[image]")
    assert not is_sequence
    assert is_link
    assert htype == "image"

    is_sequence, is_link, htype = parse_complex_htype("link[sequence[image]]")
    assert is_sequence
    assert is_link
    assert htype == "image"

    is_sequence, is_link, htype = parse_complex_htype("sequence[link[video]]")
    assert is_sequence
    assert is_link
    assert htype == "video"

    bad_inputs = [
        "random[image]",
        "sequence[random[image]]",
        "link[random[image]]",
        "link(image)",
        "sequence(image)",
        "link[sequence(image)]",
        "sequence[link(image)]",
        "link[sequence[image[uint8]]]",
        "sequence[link[image[uint8]]]",
    ]
    for bad_input in bad_inputs:
        with pytest.raises(TensorMetaInvalidHtype):
            parse_complex_htype(bad_input)


def test_link_creds(request):
    link_creds = LinkCreds()
    link_creds.add_creds_key("abc")
    link_creds.add_creds_key("def")

    with pytest.raises(ValueError):
        link_creds.add_creds_key("abc")

    link_creds.populate_creds("abc", {})
    link_creds.populate_creds("def", {})

    with pytest.raises(KeyError):
        link_creds.populate_creds("ghi", {})

    assert link_creds.get_encoding("ENV") == 0
    assert link_creds.get_encoding(None) == 0
    with pytest.raises(ValueError):
        link_creds.get_encoding(None, "s3://my_bucket/my_key")
    assert link_creds.get_encoding("abc") == 1
    assert link_creds.get_encoding("def") == 2
    with pytest.raises(ValueError):
        link_creds.get_encoding("ghi")

    assert link_creds.get_creds_key(0) is None
    assert link_creds.get_creds_key(1) == "abc"
    assert link_creds.get_creds_key(2) == "def"
    with pytest.raises(KeyError):
        link_creds.get_creds_key(3)

    assert len(link_creds) == 2
    assert link_creds.missing_keys == []

    link_creds.add_creds_key("ghi")
    assert link_creds.missing_keys == ["ghi"]

    with pytest.raises(KeyError):
        link_creds.get_storage_provider("xyz", "s3")

    with pytest.raises(ValueError):
        link_creds.get_storage_provider("ghi", "s3")

    if is_opt_true(request, GCS_OPT):
        assert isinstance(link_creds.get_storage_provider("def", "gcs"), GCSProvider)
        assert isinstance(link_creds.get_storage_provider("def", "gcs"), GCSProvider)
        assert isinstance(link_creds.get_storage_provider("ENV", "gcs"), GCSProvider)
    if is_opt_true(request, S3_OPT):
        assert isinstance(link_creds.get_storage_provider("abc", "s3"), S3Provider)
        assert isinstance(link_creds.get_storage_provider("abc", "s3"), S3Provider)
        assert isinstance(link_creds.get_storage_provider(None, "s3"), S3Provider)

    pickled = pickle.dumps(link_creds)
    unpickled_link_creds = pickle.loads(pickled)

    assert len(unpickled_link_creds) == 3
    assert unpickled_link_creds.get_creds_key(0) is None
    assert unpickled_link_creds.get_creds_key(1) == "abc"
    assert unpickled_link_creds.get_creds_key(2) == "def"
    assert unpickled_link_creds.missing_keys == ["ghi"]

    bts = link_creds.tobytes()
    assert len(bts) == link_creds.nbytes

    from_buffer_link_creds = LinkCreds.frombuffer(bts)
    assert len(from_buffer_link_creds) == 3
    assert from_buffer_link_creds.missing_keys == ["abc", "def", "ghi"]


def test_creds_encoder():
    enc = CredsEncoder()

    enc.register_samples((3,), 5)
    enc.register_samples((0,), 2)
    enc.register_samples((1,), 3)

    for i in range(5):
        assert enc[i] == (3,)
    for i in range(5, 7):
        assert enc[i] == (0,)
    for i in range(7, 10):
        assert enc[i] == (1,)

    data = enc.tobytes()
    assert len(data) == enc.nbytes

    dec = CredsEncoder.frombuffer(data)
    for i in range(5):
        assert dec[i] == (3,)
    for i in range(5, 7):
        assert dec[i] == (0,)
    for i in range(7, 10):
        assert dec[i] == (1,)


def test_add_populate_creds(local_ds_generator):
    local_ds = local_ds_generator()
    with local_ds as ds:
        ds.add_creds_key("my_s3_key")
        ds.add_creds_key("my_gcs_key")
        ds.populate_creds("my_s3_key", {})
        ds.populate_creds("my_gcs_key", {})

        assert ds.link_creds.creds_keys == ["my_s3_key", "my_gcs_key"]
        assert ds.link_creds.creds_mapping == {"my_s3_key": 1, "my_gcs_key": 2}
        assert ds.link_creds.creds_dict == {"my_s3_key": {}, "my_gcs_key": {}}

    ds = local_ds_generator()
    assert ds.link_creds.creds_keys == ["my_s3_key", "my_gcs_key"]
    assert ds.link_creds.creds_mapping == {"my_s3_key": 1, "my_gcs_key": 2}
    assert ds.link_creds.creds_dict == {}


def test_none_used_key(local_ds_generator, cat_path):
    local_ds = local_ds_generator()
    with local_ds as ds:
        ds.create_tensor("xyz", htype="link[image]")
        ds.add_creds_key("my_s3_key")
        ds.populate_creds("my_s3_key", {})
        ds.xyz.append(hub.link(cat_path))
        assert ds.link_creds.used_creds_keys == set()
        ds.xyz.append(hub.link(cat_path, "ENV"))
        assert ds.link_creds.used_creds_keys == set()
        ds.xyz.append(hub.link(cat_path, "my_s3_key"))
        assert ds.link_creds.used_creds_keys == {"my_s3_key"}

    ds = local_ds_generator()
    assert ds.link_creds.used_creds_keys == {"my_s3_key"}


@pytest.mark.parametrize("create_shape_tensor", [True, False])
@pytest.mark.parametrize("verify", [True, False])
def test_basic(local_ds_generator, cat_path, flower_path, create_shape_tensor, verify):
    local_ds = local_ds_generator()
    with local_ds as ds:
        ds.create_tensor(
            "linked_images",
            htype="link[image]",
            create_shape_tensor=create_shape_tensor,
            verify=verify,
            sample_compression="jpeg",
        )
        with pytest.raises(TypeError):
            ds.linked_images.append(np.ones((100, 100, 3)))

        for _ in range(10):
            sample = hub.link(flower_path)
            ds.linked_images.append(sample)

        ds.linked_images.append(None)

        for i in range(0, 10, 2):
            sample = hub.link(cat_path)
            ds.linked_images[i] = sample

        assert len(ds.linked_images) == 11

        ds.create_tensor(
            "linked_images_2",
            htype="link[image]",
            create_shape_tensor=create_shape_tensor,
            verify=verify,
            sample_compression="jpeg",
        )
        ds.linked_images_2.extend(ds.linked_images)
        assert len(ds.linked_images_2) == 11
        for i in range(10):
            shape_target = (900, 900, 3) if i % 2 == 0 else (513, 464, 4)
            assert ds.linked_images[i].shape == shape_target
            assert ds.linked_images[i].numpy().shape == shape_target
            assert ds.linked_images_2[i].shape == shape_target
            assert ds.linked_images_2[i].numpy().shape == shape_target

        assert ds.linked_images[10].shape == (0,)
        np.testing.assert_array_equal(ds.linked_images[10].numpy(), np.ones((0,)))
        assert ds.linked_images_2[10].shape == (0,)
        np.testing.assert_array_equal(ds.linked_images_2[10].numpy(), np.ones((0,)))

    ds.commit()
    view = ds[:5]
    view.save_view(optimize=True)
    view2 = ds.get_views()[0].load()
    view1_np = view.linked_images.numpy(aslist=True)
    view2_np = view2.linked_images.numpy(aslist=True)
    assert len(view1_np) == len(view2_np)
    for v1, v2 in zip(view1_np, view2_np):
        np.testing.assert_array_equal(v1, v2)

    # checking persistence
    ds = local_ds_generator()
    for i in range(10):
        shape_target = (900, 900, 3) if i % 2 == 0 else (513, 464, 4)
        assert ds.linked_images[i].shape == shape_target
        assert ds.linked_images[i].numpy().shape == shape_target
        assert ds.linked_images_2[i].shape == shape_target
        assert ds.linked_images_2[i].numpy().shape == shape_target


def test_jwt_link(local_ds):
    with local_ds as ds:
        ds.create_tensor("img", htype="link[image]", create_shape_tensor=False)
        auth = HubBackendClient().auth_header
        my_jwt = {"Authorization": auth}
        ds.add_creds_key("my_jwt_key")
        ds.populate_creds("my_jwt_key", my_jwt)
        img_url = "https://app-dev.activeloop.dev/api/org/tim4/storage/image"
        for _ in range(3):
            ds.img.append(hub.link(img_url, creds_key="my_jwt_key"))

        for i in range(3):
            assert ds.img[i].shape == (50, 50, 4)
            assert ds.img[i].numpy().shape == (50, 50, 4)

        my_incorrect_jwt = {"Authorization": "12345"}
        ds.populate_creds("my_jwt_key", my_incorrect_jwt)
        with pytest.raises(UnableToReadFromUrlError):
            ds.img[0].numpy()

        with pytest.raises(UnableToReadFromUrlError):
            ds.img[0].shape


@pytest.mark.parametrize("create_shape_tensor", [True, False])
@pytest.mark.parametrize("verify", [True, False])
@pytest.mark.skipif(
    os.name == "nt" and sys.version_info < (3, 7), reason="requires python 3.7 or above"
)
def test_video(request, local_ds_generator, create_shape_tensor, verify):
    local_ds = local_ds_generator()
    with local_ds as ds:
        ds.create_tensor(
            "linked_videos",
            htype="link[video]",
            create_shape_tensor=create_shape_tensor,
            verify=verify,
        )
        for _ in range(3):
            sample = hub.link(
                "http://commondatastorage.googleapis.com/gtv-videos-bucket/sample/ForBiggerJoyrides.mp4"
            )
            ds.linked_videos.append(sample)

        assert len(ds.linked_videos) == 3
        for i in range(3):
            assert ds.linked_videos[i].shape == (361, 720, 1280, 3)
            assert ds.linked_videos[i][:5].numpy().shape == (5, 720, 1280, 3)

        if is_opt_true(request, GCS_OPT):
            sample = hub.link(
                "gcs://gtv-videos-bucket/sample/ForBiggerJoyrides.mp4", creds_key="ENV"
            )
            ds.linked_videos.append(sample)
            assert len(ds.linked_videos) == 4
            assert ds.linked_videos[3].shape == (361, 720, 1280, 3)
    # checking persistence
    ds = local_ds_generator()
    for i in range(3):
        assert ds.linked_videos[i].shape == (361, 720, 1280, 3)

    if is_opt_true(request, GCS_OPT):
        assert len(ds.linked_videos) == 4
        assert ds.linked_videos[3].shape == (361, 720, 1280, 3)


def test_complex_creds(local_ds_generator):
    local_ds = local_ds_generator()
    with local_ds as ds:
        ds.create_tensor(
            "link",
            htype="link[image]",
            verify=False,
            create_shape_tensor=False,
            create_sample_info_tensor=False,
        )
        ds.create_tensor("xyz")
        ds.add_creds_key("my_first_key")
        ds.add_creds_key("my_second_key")

        assert ds.get_creds_keys() == ["my_first_key", "my_second_key"]

        ds.populate_creds("my_first_key", {})
        ds.populate_creds("my_second_key", {})
        for i in range(10):
            creds_key = "my_first_key" if i % 2 == 0 else "my_second_key"
            sample = hub.link("https://picsum.photos/200/300", creds_key=creds_key)
            ds.link.append(sample)
            ds.xyz.append(i)

        with pytest.raises(ValueError):
            ds.link._linked_sample()

        with pytest.raises(ValueError):
            ds.xyz[0]._linked_sample()

        linked_sample = ds.link[0]._linked_sample()
        assert linked_sample.path == "https://picsum.photos/200/300"
        assert linked_sample.creds_key == "my_first_key"

        for i in range(10, 15):
            sample = hub.link("https://picsum.photos/200/300")
            ds.link.append(sample)
            ds.xyz.append(i)

        for i in range(10):
            enc_creds = 1 if i % 2 == 0 else 2
            assert ds.link.chunk_engine.creds_encoder[i][0] == enc_creds

        for i in range(10, 15):
            assert ds.link.chunk_engine.creds_encoder[i][0] == 0

        for i in range(15):
            assert ds.xyz[i].numpy() == i
            assert ds.link[i].numpy().shape == (300, 200, 3)

    ds = local_ds_generator()
    for i in range(10):
        enc_creds = 1 if i % 2 == 0 else 2
        assert ds.link.chunk_engine.creds_encoder[i][0] == enc_creds

    for i in range(10, 15):
        assert ds.link.chunk_engine.creds_encoder[i][0] == 0

    for i in range(15):
        assert ds.xyz[i].numpy() == i

    with pytest.raises(ValueError):
        ds.link[0].numpy().shape


@hub.compute
def identity(sample_in, samples_out):
    samples_out.linked_images.append(sample_in.linked_images)


def test_transform(local_ds, cat_path, flower_path):
    data_in = hub.dataset("./test/link_transform", overwrite=True)
    with data_in as ds:
        ds.create_tensor(
            "linked_images",
            htype="link[image]",
            create_shape_tensor=True,
            verify=True,
            sample_compression="jpeg",
        )
        for i in range(10):
            sample = hub.link(cat_path) if i % 2 == 0 else hub.link(flower_path)
            ds.linked_images.append(sample)

    data_out = local_ds
    with data_out as ds:
        ds.create_tensor(
            "linked_images",
            htype="link[image]",
            create_shape_tensor=True,
            verify=True,
            sample_compression="jpeg",
        )

    identity().eval(data_in, data_out, num_workers=2)
    assert len(data_out.linked_images) == 10
    for i in range(10):
        shape_target = (900, 900, 3) if i % 2 == 0 else (513, 464, 4)
        assert ds.linked_images[i].shape == shape_target
        assert ds.linked_images[i].numpy().shape == shape_target

    data_in.delete()


@hub.compute
def transform_path_link(sample_in, samples_out):
    samples_out.images.append(hub.link(sample_in))


def check_transformed_ds(ds):
    assert ds.images[0].numpy().shape == ds.images[0].shape == (900, 900, 3)
    assert ds.images[1].numpy().shape == ds.images[1].shape == (513, 464, 4)


def test_transform_2(local_ds_generator, cat_path, flower_path):
    ds = local_ds_generator()
    with ds:
        ds.create_tensor("images", htype="link[image]")

    transform_path_link().eval([cat_path, flower_path], ds)

    check_transformed_ds(ds)
    ds = local_ds_generator()
    check_transformed_ds(ds)


def test_link_managed(hub_cloud_ds_generator, cat_path):
    key_name = "CREDS_MANAGEMENT_TEST"
    with hub_cloud_ds_generator() as ds:
        ds.create_tensor(
            "img",
            htype="link[image]",
            verify=False,
            create_shape_tensor=False,
            create_sample_info_tensor=False,
        )
        ds.add_creds_key(key_name, managed=True)
        assert key_name in ds.link_creds.creds_dict
        assert key_name in ds.link_creds.managed_creds_keys
        assert key_name not in ds.link_creds.used_creds_keys

        ds.img.append(hub.link(cat_path, creds_key=key_name))
        assert key_name in ds.link_creds.used_creds_keys

    ds = hub_cloud_ds_generator()
    assert key_name in ds.link_creds.creds_dict
    assert key_name in ds.link_creds.managed_creds_keys
    assert key_name in ds.link_creds.used_creds_keys

    shape_target = (900, 900, 3)
    assert ds.img[0].shape == shape_target
    assert ds.img[0].numpy().shape == shape_target

    with pytest.raises(ValueError):
        # managed creds_key can't be updated
        ds.update_creds_key(key_name, "something_else")

    with pytest.raises(KeyError):
        ds.change_creds_management("random_key", False)

    # this is a no-op
    ds.change_creds_management(key_name, True)

    # no longer managed
    ds.change_creds_management(key_name, False)

    ds = hub_cloud_ds_generator()
    with pytest.raises(ValueError):
        ds.img[0].numpy()

    ds.populate_creds(key_name, {})
    assert ds.img[0].shape == shape_target
    assert ds.img[0].numpy().shape == shape_target

    ds = hub_cloud_ds_generator()
    ds.change_creds_management(key_name, True)
    assert ds.img[0].shape == shape_target
    assert ds.img[0].numpy().shape == shape_target

    new_key = "some_random_key"
    with pytest.raises(ManagedCredentialsNotFoundError):
        ds.add_creds_key(new_key, managed=True)

    # even after failure one can simply add a new key, setting managed to False
    ds.add_creds_key(new_key)


def test_link_ready(local_ds_generator, cat_path):
    with local_ds_generator() as ds:
        ds.create_tensor(
            "img",
            htype="link[image]",
            verify=False,
            create_shape_tensor=False,
            create_sample_info_tensor=False,
        )
        ds.add_creds_key("def")
        ds.add_creds_key("abc")
        ds.populate_creds("abc", {})
        ds.img.append(hub.link(cat_path, creds_key="abc"))

    ds = local_ds_generator()
    with pytest.raises(ValueError):
        ds.img[0].numpy()
    ds.populate_creds("abc", {})
    assert ds.img[0].numpy().shape == (900, 900, 3)
    with pytest.raises(KeyError):
        ds.update_creds_key("xyz", "ghi")
    with pytest.raises(ValueError):
        ds.update_creds_key("abc", "def")
    ds.update_creds_key("abc", "new")
    assert ds.img[0].numpy().shape == (900, 900, 3)
    ds = local_ds_generator()
    with pytest.raises(ValueError):
        ds.img[0].numpy()
    ds.populate_creds("new", {})
    assert ds.img[0].numpy().shape == (900, 900, 3)


@pytest.mark.parametrize("create_shape_tensor", [True, False])
@pytest.mark.parametrize("verify", [True, False])
def test_basic_sequence(local_ds, cat_path, flower_path, create_shape_tensor, verify):
    pass



import subprocess
import pytest
from hub.cli.commands import login
from click.testing import CliRunner
from hub.client.client import HubBackendClient
from hub.client.utils import (
    write_token,
    read_token,
    remove_token,
)


def test_client_requests(hub_cloud_dev_credentials):
    username, password = hub_cloud_dev_credentials

    hub_client = HubBackendClient()
    hub_client.request_auth_token(username, password)
    with pytest.raises(Exception):
        # request will fail as username already exists
        hub_client.send_register_request("activeloop", "abc@d.com", "notactualpassword")


def test_client_utils():
    write_token("abcdefgh")
    assert read_token() == "abcdefgh"
    remove_token()
    assert read_token() is None


def test_client_workspace_organizations(hub_cloud_dev_credentials):
    username, password = hub_cloud_dev_credentials
    hub_client = HubBackendClient()

    assert hub_client.get_user_organizations() == ["public"]
    token = hub_client.request_auth_token(username, password)
    runner = CliRunner()
    runner.invoke(login, f"-u {username} -p {password}")
    hub_client = HubBackendClient()
    assert username in hub_client.get_user_organizations()
    assert "public" in hub_client.get_user_organizations()

    datasets = subprocess.check_output(
        ["activeloop", "list-datasets", "--workspace", "activeloop"]
    )
    assert "You are not a member of organization" in str(datasets)
    datasets = subprocess.check_output(
        ["activeloop", "list-datasets", "--workspace", "test"]
    )
    assert "You are not a member of organization" not in str(datasets)



from fractions import Fraction

from PIL import Image  # type: ignore
import PIL.ExifTags  # type: ignore


_LOOKUPS = {
    "MeteringMode": (
        "Undefined",
        "Average",
        "Center-weighted average",
        "Spot",
        "Multi-spot",
        "Multi-segment",
        "Partial",
    ),
    "ExposureProgram": (
        "Undefined",
        "Manual",
        "Program AE",
        "Aperture-priority AE",
        "Shutter speed priority AE",
        "Creative (Slow speed)",
        "Action (High speed)",
        "Portrait ",
        "Landscape",
        "Bulb",
    ),
    "ResolutionUnit": ("", "Undefined", "Inches", "Centimetres"),
    "Orientation": (
        "",
        "Horizontal",
        "Mirror horizontal",
        "Rotate 180",
        "Mirror vertical",
        "Mirror horizontal and rotate 270 CW",
        "Rotate 90 CW",
        "Mirror horizontal and rotate 90 CW",
        "Rotate 270 CW",
    ),
}


def getexif(image: Image):
    raw_exif = image._getexif()
    if not raw_exif:
        return {}
    exif = {}
    for k, v in raw_exif.items():
        tag = k
        k = PIL.ExifTags.TAGS.get(k, k)
        v = _process_exif_value(k, v)
        if v is None:
            continue
        exif[k] = v
    return exif


def _process_exif_value(k: str, v):
    if hasattr(v, "numerator"):  # Rational
        v = v.numerator / v.denominator if v.denominator else 0
        if k == "ExposureTime":
            v = float(Fraction(v).limit_denominator(8000))
        elif k in ("XResolution", "YResolution"):
            try:
                v = int(v)
            except ValueError:  # nan
                pass
        return v
    elif k in _LOOKUPS:
        return _LOOKUPS[k][v]
    elif isinstance(v, bytes):
        return str(v) if len(v) < 16 else "<bytes>"
    elif isinstance(v, (list, tuple)):
        return type(v)(_process_exif_value("_", x) for x in v)
    elif isinstance(v, dict):
        return {k: _process_exif_value("_", x) for k, x in v.items()}
    return str(v)



from hub.core.storage.gcs import GCSProvider
from hub.util.agreement import handle_dataset_agreements
from hub.util.cache_chain import generate_chain
from hub.constants import LOCAL_CACHE_PREFIX, MB
from hub.util.exceptions import AgreementNotAcceptedError
from hub.util.tag import process_hub_path
from typing import Optional
from hub.core.storage.provider import StorageProvider
import os
from hub.core.storage import (
    LocalProvider,
    S3Provider,
    MemoryProvider,
    GDriveProvider,
)
from hub.client.client import HubBackendClient
import posixpath
from hub.constants import DEFAULT_READONLY


def storage_provider_from_path(
    path: str,
    creds: Optional[dict],
    read_only: bool = False,
    token: Optional[str] = None,
    is_hub_path: bool = False,
):
    """Construct a StorageProvider given a path.

    Arguments:
        path (str): The full path to the Dataset.
        creds (dict): A dictionary containing credentials used to access the dataset at the url.
            This takes precedence over credentials present in the environment. Only used when url is provided. Currently only works with s3 urls.
        read_only (bool): Opens dataset in read only mode if this is passed as True. Defaults to False.
        token (str): token for authentication into activeloop.
        is_hub_path (bool): whether the path points to a hub dataset.

    Returns:
        If given a path starting with s3:// returns the S3Provider.
        If given a path starting with gcp:// or gcs:// returns the GCPProvider.
        If given a path starting with gdrive:// returns the GDriveProvider
        If given a path starting with mem:// returns the MemoryProvider.
        If given a path starting with hub:// returns the underlying cloud Provider.
        If given a valid local path, returns the LocalProvider.

    Raises:
        ValueError: If the given path is a local path to a file.
    """
    if creds is None:
        creds = {}
    if path.startswith("s3://"):
        key = creds.get("aws_access_key_id")
        secret = creds.get("aws_secret_access_key")
        session_token = creds.get("aws_session_token")
        endpoint_url = creds.get("endpoint_url")
        region = creds.get("aws_region")
        profile = creds.get("profile_name")
        storage: StorageProvider = S3Provider(
            path,
            key,
            secret,
            session_token,
            endpoint_url,
            region,
            profile_name=profile,
            token=token,
        )
    elif (
        path.startswith("gcp://")
        or path.startswith("gcs://")
        or path.startswith("gs://")
    ):
        storage = GCSProvider(path, creds)
    elif path.startswith("gdrive://"):
        storage = GDriveProvider(path, creds)
    elif path.startswith("mem://"):
        storage = MemoryProvider(path)
    elif path.startswith("hub://"):
        storage = storage_provider_from_hub_path(path, read_only, token=token)
    else:
        if not os.path.exists(path) or os.path.isdir(path):
            storage = LocalProvider(path)
        else:
            raise ValueError(f"Local path {path} must be a path to a local directory")
    if not storage._is_hub_path:
        storage._is_hub_path = is_hub_path

    if read_only:
        storage.enable_readonly()
    return storage


def storage_provider_from_hub_path(
    path: str, read_only: bool = False, token: str = None
):
    path, org_id, ds_name, subdir = process_hub_path(path)
    client = HubBackendClient(token=token)

    mode = "r" if read_only else None

    # this will give the proper url (s3, gcs, etc) and corresponding creds, depending on where the dataset is stored.
    try:
        url, creds, mode, expiration = client.get_dataset_credentials(
            org_id, ds_name, mode=mode
        )
    except AgreementNotAcceptedError as e:
        handle_dataset_agreements(client, e.agreements, org_id, ds_name)
        url, creds, mode, expiration = client.get_dataset_credentials(
            org_id, ds_name, mode=mode
        )

    if mode == "r":
        read_only = True
        if read_only is None and not DEFAULT_READONLY:
            # warns user about automatic mode change
            print(
                "Opening dataset in read-only mode as you don't have write permissions."
            )

    if read_only is None:
        read_only = DEFAULT_READONLY

    url = posixpath.join(url, subdir)

    storage = storage_provider_from_path(
        path=url, creds=creds, read_only=read_only, is_hub_path=True
    )
    storage._set_hub_creds_info(path, expiration)
    return storage


def get_storage_and_cache_chain(
    path, read_only, creds, token, memory_cache_size, local_cache_size
):
    """
    Returns storage provider and cache chain for a given path, according to arguments passed.

    Args:
        path (str): The full path to the Dataset.
        creds (dict): A dictionary containing credentials used to access the dataset at the url.
            This takes precedence over credentials present in the environment. Only used when url is provided. Currently only works with s3 urls.
        read_only (bool): Opens dataset in read only mode if this is passed as True. Defaults to False.
        token (str): token for authentication into activeloop
        memory_cache_size (int): The size of the in-memory cache to use.
        local_cache_size (int): The size of the local cache to use.

    Returns:
        A tuple of the storage provider and the storage chain.
    """
    storage = storage_provider_from_path(
        path=path,
        creds=creds,
        read_only=read_only,
        token=token,
    )
    memory_cache_size_bytes = memory_cache_size * MB
    local_cache_size_bytes = local_cache_size * MB
    storage_chain = generate_chain(
        storage, memory_cache_size_bytes, local_cache_size_bytes, path
    )
    return storage, storage_chain


def get_local_storage_path(path: str, prefix: str):
    local_cache_name = path.replace("://", "_")
    local_cache_name = local_cache_name.replace("/", "_")
    local_cache_name = local_cache_name.replace("\\", "_")
    return f"{prefix}/{local_cache_name}"


def get_pytorch_local_storage(dataset):
    """Returns a local storage provider for a given dataset to be used for Pytorch iteration"""
    local_cache_name: str = f"{dataset.path}_pytorch"
    local_cache_path = get_local_storage_path(local_cache_name, LOCAL_CACHE_PREFIX)
    return LocalProvider(local_cache_path)



from typing import Set, List
from hub.core.storage.hub_memory_object import HubMemoryObject


class CommitDiff(HubMemoryObject):
    """Stores set of diffs stored for a particular tensor in a commit."""

    def __init__(self, first_index=0, created=False) -> None:
        self.is_dirty = created  # only put as dirty during init if created
        self.created = created
        self.data_added: List[int] = [first_index, first_index]
        self.data_updated: Set[int] = set()
        self.data_deleted: Set[int] = set()
        self.info_updated = False
        self.cleared = False

        # this is stored for in place transforms in which we no longer need to considered older diffs about added/updated data
        self.data_transformed = False

    def tobytes(self) -> bytes:
        """Returns bytes representation of the commit diff

        The format stores the following information in order:
        1. The first byte is a boolean value indicating whether the tensor was created in the commit or not.
        2. The second byte is a boolean value indicating whether the info has been updated or not.
        3. The third byte is a boolean value indicating whether the data has been transformed using an inplace transform or not.
        4. The next 8 + 8 bytes are the two elements of the data_added list.
        5. The next 8 bytes are the number of elements in the data_updated set, let's call this m.
        6. The next 8 * m bytes are the elements of the data_updated set.
        7. The next byte is a boolean value indicating whether the tensor was cleared in the commit or not.
        8. The next 8 bytes are the number of elements in the data_deleted set, let's call this n.
        9. The next 8 * n bytes are the elements of the data_deleted set.
        """
        return b"".join(
            [
                self.created.to_bytes(1, "big"),
                self.info_updated.to_bytes(1, "big"),
                self.data_transformed.to_bytes(1, "big"),
                self.data_added[0].to_bytes(8, "big"),
                self.data_added[1].to_bytes(8, "big"),
                len(self.data_updated).to_bytes(8, "big"),
                *(idx.to_bytes(8, "big") for idx in self.data_updated),
                self.cleared.to_bytes(1, "big"),
                len(self.data_deleted).to_bytes(8, "big"),
                *(idx.to_bytes(8, "big") for idx in self.data_deleted),
            ]
        )

    @classmethod
    def frombuffer(cls, data: bytes) -> "CommitDiff":
        """Creates a CommitDiff object from bytes"""
        commit_diff = cls()

        commit_diff.created = bool(int.from_bytes(data[:1], "big"))
        commit_diff.info_updated = bool(int.from_bytes(data[1:2], "big"))
        commit_diff.data_transformed = bool(int.from_bytes(data[2:3], "big"))
        commit_diff.data_added = [
            int.from_bytes(data[3:11], "big"),
            int.from_bytes(data[11:19], "big"),
        ]
        num_updates = int.from_bytes(data[19:27], "big")
        commit_diff.data_updated = {
            int.from_bytes(data[27 + i * 8 : 35 + i * 8], "big")
            for i in range(num_updates)
        }
        pos = 35 + (num_updates - 1) * 8
        commit_diff.cleared = bool(int.from_bytes(data[pos : pos + 1], "big"))
        commit_diff.is_dirty = False
        pos += 1
        if len(data) > pos:
            num_deletes = int.from_bytes(data[pos : pos + 8], "big")
            pos += 8
            commit_diff.data_deleted = {
                int.from_bytes(data[pos + i * 8 : pos + i * 8 + 8], "big")
                for i in range(num_deletes)
            }
        else:
            commit_diff.data_deleted = set()
        return commit_diff

    @property
    def nbytes(self):
        """Returns number of bytes required to store the commit diff"""
        return 36 + 8 * (len(self.data_updated) + len(self.data_deleted))

    @property
    def num_samples_added(self) -> int:
        """Returns number of samples added"""
        return self.data_added[1] - self.data_added[0]

    def modify_info(self) -> None:
        """Stores information that the info has changed"""
        self.info_updated = True
        self.is_dirty = True

    def add_data(self, count: int) -> None:
        """Adds new indexes to data added"""
        self.data_added[1] += count
        self.is_dirty = True

    def update_data(self, global_index: int) -> None:
        """Adds new indexes to data updated"""
        global_index = self.translate_index(global_index)
        if global_index not in self.data_added:
            self.data_updated.add(global_index)
            self.is_dirty = True

    def clear_data(self):
        """Clears data"""
        self.data_added = [0, 0]
        self.data_updated = set()
        self.data_deleted = set()
        self.info_updated = False
        self.cleared = True
        self.is_dirty = True

    def transform_data(self) -> None:
        """Stores information that the data has been transformed using an inplace transform."""
        self.data_transformed = True
        self.is_dirty = True

    def pop(self, index) -> None:
        index = self.translate_index(index)
        if index not in range(*self.data_added):
            self.data_deleted.add(index)
            self.data_added[0] -= 1
        self.data_added[1] -= 1

        if index in self.data_updated:
            self.data_updated.remove(index)

        self.data_updated = {
            idx - 1 if idx > index else idx for idx in self.data_updated
        }
        self.is_dirty = True

    def translate_index(self, index):
        if not self.data_deleted:
            return index
        offset = sum(i < index for i in self.data_deleted)
        return index + offset



import hub
from hub.core.linked_chunk_engine import LinkedChunkEngine
from hub.core.storage.lru_cache import LRUCache
from hub.util.invalid_view_op import invalid_view_op
from hub.core.version_control.commit_chunk_set import CommitChunkSet
from hub.core.version_control.commit_diff import CommitDiff
from hub.core.chunk.base_chunk import InputSample
import numpy as np
from typing import Dict, List, Sequence, Union, Optional, Tuple, Any, Callable
from functools import reduce, partial
from hub.core.index import Index, IndexEntry
from hub.core.meta.tensor_meta import TensorMeta
from hub.core.storage import StorageProvider
from hub.core.chunk_engine import ChunkEngine
from hub.core.compression import _read_timestamps
from hub.core.tensor_link import get_link_transform
from hub.api.info import Info, load_info
from hub.util.keys import (
    get_chunk_id_encoder_key,
    get_chunk_key,
    get_tensor_commit_chunk_set_key,
    get_tensor_commit_diff_key,
    get_tensor_meta_key,
    get_tensor_tile_encoder_key,
    get_sequence_encoder_key,
    tensor_exists,
    get_tensor_info_key,
    get_sample_id_tensor_key,
    get_sample_info_tensor_key,
    get_sample_shape_tensor_key,
)
from hub.util.modified import get_modified_indexes
from hub.util.class_label import convert_to_text
from hub.util.shape_interval import ShapeInterval
from hub.util.exceptions import (
    TensorDoesNotExistError,
    InvalidKeyTypeError,
    TensorAlreadyExistsError,
)
from hub.hooks import dataset_read, dataset_written
from hub.util.pretty_print import (
    summary_tensor,
)
from hub.constants import FIRST_COMMIT_ID, _NO_LINK_UPDATE, UNSPECIFIED


from hub.util.version_control import auto_checkout
from hub.util.video import normalize_index

from hub.compression import get_compression_type, VIDEO_COMPRESSION
from hub.util.notebook import is_jupyter, video_html, is_colab
from hub.util.point_cloud import (
    POINT_CLOUD_FIELD_NAME_TO_TYPESTR,
    cast_point_cloud_array_to_proper_dtype,
)
import warnings
import webbrowser


def create_tensor(
    key: str,
    storage: StorageProvider,
    htype: str,
    sample_compression: str,
    chunk_compression: str,
    version_state: Dict[str, Any],
    **kwargs,
):
    """If a tensor does not exist, create a new one with the provided meta.

    Args:
        key (str): Key for where the chunks, index_meta, and tensor_meta will be located in `storage` relative to it's root.
        storage (StorageProvider): StorageProvider that all tensor data is written to.
        htype (str): Htype is how the default tensor metadata is defined.
        sample_compression (str): All samples will be compressed in the provided format. If `None`, samples are uncompressed.
        chunk_compression (str): All chunks will be compressed in the provided format. If `None`, chunks are uncompressed.
        version_state (Dict[str, Any]): The version state of the dataset, includes commit_id, commit_node, branch, branch_commit_map and commit_node_map.
        **kwargs: `htype` defaults can be overridden by passing any of the compatible parameters.
            To see all `htype`s and their correspondent arguments, check out `hub/htypes.py`.

    Raises:
        TensorAlreadyExistsError: If a tensor defined with `key` already exists.
    """

    commit_id = version_state["commit_id"]
    if tensor_exists(key, storage, commit_id):
        raise TensorAlreadyExistsError(key)

    meta_key = get_tensor_meta_key(key, commit_id)
    meta = TensorMeta(
        htype=htype,
        sample_compression=sample_compression,
        chunk_compression=chunk_compression,
        **kwargs,
    )
    storage[meta_key] = meta  # type: ignore

    if commit_id != FIRST_COMMIT_ID:
        cset_key = get_tensor_commit_chunk_set_key(key, commit_id)
        cset = CommitChunkSet()
        storage[cset_key] = cset  # type: ignore

    diff_key = get_tensor_commit_diff_key(key, commit_id)
    diff = CommitDiff(created=True)
    storage[diff_key] = diff  # type: ignore


def delete_tensor(key: str, dataset):
    """Delete tensor from storage.

    Args:
        key (str): Key for where the chunks, index_meta, and tensor_meta will be located in `storage` relative to it's root.
        dataset (Dataset): Dataset that the tensor is located in.

    Raises:
        TensorDoesNotExistError: If no tensor with `key` exists and a `tensor_meta` was not provided.
    """
    storage = dataset.storage
    version_state = dataset.version_state
    tensor = Tensor(key, dataset)
    chunk_engine: ChunkEngine = tensor.chunk_engine
    enc = chunk_engine.chunk_id_encoder
    n_chunks = chunk_engine.num_chunks
    chunk_names = [enc.get_name_for_chunk(i) for i in range(n_chunks)]
    chunk_keys = [
        get_chunk_key(key, chunk_name, version_state["commit_id"])
        for chunk_name in chunk_names
    ]
    for chunk_key in chunk_keys:
        try:
            del storage[chunk_key]
        except KeyError:
            pass

    commit_id = version_state["commit_id"]
    meta_key = get_tensor_meta_key(key, commit_id)
    try:
        del storage[meta_key]
    except KeyError:
        pass

    info_key = get_tensor_info_key(key, commit_id)
    try:
        del storage[info_key]
    except KeyError:
        pass

    diff_key = get_tensor_commit_diff_key(key, commit_id)
    try:
        del storage[diff_key]
    except KeyError:
        pass

    chunk_id_encoder_key = get_chunk_id_encoder_key(key, commit_id)
    try:
        del storage[chunk_id_encoder_key]
    except KeyError:
        pass

    tile_encoder_key = get_tensor_tile_encoder_key(key, commit_id)
    try:
        del storage[tile_encoder_key]
    except KeyError:
        pass

    seq_encoder_key = get_sequence_encoder_key(key, commit_id)
    try:
        del storage[seq_encoder_key]
    except KeyError:
        pass


def _inplace_op(f):
    op = f.__name__

    def inner(tensor, other):
        tensor._write_initialization()
        tensor.chunk_engine.update(
            tensor.index,
            other,
            op,
            link_callback=tensor._update_links if tensor.meta.links else None,
        )
        if not tensor.index.is_trivial():
            tensor._skip_next_setitem = True
        return tensor

    return inner


class Tensor:
    def __init__(
        self,
        key: str,
        dataset,
        index: Optional[Index] = None,
        is_iteration: bool = False,
        chunk_engine: Optional[ChunkEngine] = None,
    ):
        """Initializes a new tensor.

        Args:
            key (str): The internal identifier for this tensor.
            dataset (Dataset): The dataset that this tensor is located in.
            index: The Index object restricting the view of this tensor.
                Can be an int, slice, or (used internally) an Index object.
            is_iteration (bool): If this tensor is being used as an iterator.
            chunk_engine (ChunkEngine, optional): The underlying chunk_engine for the tensor.

        Raises:
            TensorDoesNotExistError: If no tensor with ``key`` exists and a ``tensor_meta`` was not provided.

        Note:
            This operation does not create a new tensor in the storage provider,
            and should normally only be performed by Hub internals.
        """
        self.key = key
        self.dataset = dataset
        self.storage: LRUCache = dataset.storage
        self.index = index or Index()
        self.version_state = dataset.version_state
        self.link_creds = dataset.link_creds
        self.is_iteration = is_iteration
        commit_id = self.version_state["commit_id"]

        if not self.is_iteration and not tensor_exists(
            self.key, self.storage, commit_id
        ):
            raise TensorDoesNotExistError(self.key)

        meta_key = get_tensor_meta_key(self.key, commit_id)
        meta = self.storage.get_hub_object(meta_key, TensorMeta)
        if chunk_engine is not None:
            self.chunk_engine = chunk_engine
        elif meta.is_link:
            self.chunk_engine = LinkedChunkEngine(
                self.key,
                self.storage,
                self.version_state,
                link_creds=dataset.link_creds,
            )
        else:
            self.chunk_engine = ChunkEngine(self.key, self.storage, self.version_state)

        if not self.pad_tensor and not self.is_iteration:
            self.index.validate(self.num_samples)

        # An optimization to skip multiple .numpy() calls when performing inplace ops on slices:
        self._skip_next_setitem = False

    @property
    def pad_tensor(self):
        return self.dataset._pad_tensors

    def _write_initialization(self):
        self.storage.check_readonly()
        # if not the head node, checkout to an auto branch that is newly created
        if auto_checkout(self.dataset):
            self.chunk_engine = self.version_state["full_tensors"][
                self.key
            ].chunk_engine

    @invalid_view_op
    def extend(
        self,
        samples: Union[np.ndarray, Sequence[InputSample], "Tensor"],
        progressbar: bool = False,
    ):

        """Extends the end of the tensor by appending multiple elements from a sequence. Accepts a sequence, a single batched numpy array,
        or a sequence of :func:`hub.read` outputs, which can be used to load files. See examples down below.

        Example:
            Numpy input:

            >>> len(tensor)
            0
            >>> tensor.extend(np.zeros((100, 28, 28, 1)))
            >>> len(tensor)
            100


            File input:

            >>> len(tensor)
            0
            >>> tensor.extend([
                    hub.read("path/to/image1"),
                    hub.read("path/to/image2"),
                ])
            >>> len(tensor)
            2


        Args:
            samples (np.ndarray, Sequence, Sequence[Sample]): The data to add to the tensor.
                The length should be equal to the number of samples to add.
            progressbar (bool): Specifies whether a progressbar should be displayed while extending.

        Raises:
            TensorDtypeMismatchError: Dtype for array must be equal to or castable to this tensor's dtype.
        """
        self._write_initialization()
        [f() for f in list(self.dataset._update_hooks.values())]
        self.chunk_engine.extend(
            samples,
            progressbar=progressbar,
            link_callback=self._append_to_links if self.meta.links else None,
        )
        dataset_written(self.dataset)

    @property
    def info(self) -> Info:
        """Returns the information about the tensor. User can set info of tensor.

        Returns:
            Info: Information about the tensor.

        Example:

            >>> # update info
            >>> ds.images.info.update(large=True, gray=False)
            >>> # get info
            >>> ds.images.info
            {'large': True, 'gray': False}

            >>> ds.images.info = {"complete": True}
            >>> ds.images.info
            {'complete': True}

        """
        commit_id = self.version_state["commit_id"]
        chunk_engine = self.chunk_engine
        if chunk_engine._info is None or chunk_engine._info_commit_id != commit_id:
            path = get_tensor_info_key(self.key, commit_id)
            chunk_engine._info = load_info(path, self.dataset, self.key)
            chunk_engine._info_commit_id = commit_id
            self.storage.register_hub_object(path, chunk_engine._info)
        return chunk_engine._info

    @info.setter
    def info(self, value):
        if isinstance(value, dict):
            info = self.info
            info.replace_with(value)
        else:
            raise TypeError("Info must be set with type Dict")

    @invalid_view_op
    def append(self, sample: InputSample):
        """Appends a single sample to the end of the tensor. Can be an array, scalar value, or the return value from :func:`hub.read`,
        which can be used to load files. See examples down below.

        Examples:
            Numpy input:

            >>> len(tensor)
            0
            >>> tensor.append(np.zeros((28, 28, 1)))
            >>> len(tensor)
            1

            File input:

            >>> len(tensor)
            0
            >>> tensor.append(hub.read("path/to/file"))
            >>> len(tensor)
            1

        Args:
            sample (InputSample): The data to append to the tensor. :class:`~hub.core.sample.Sample` is generated by :func:`hub.read`. See the above examples.
        """
        self.extend([sample], progressbar=False)

    def clear(self):
        """Deletes all samples from the tensor"""
        self.chunk_engine.clear()
        sample_id_key = get_sample_id_tensor_key(self.key)
        try:
            sample_id_tensor = Tensor(sample_id_key, self.dataset)
            sample_id_tensor.chunk_engine.clear()
            self.meta.links.clear()
            self.meta.is_dirty = True
        except TensorDoesNotExistError:
            pass

    def modified_samples(
        self, target_id: Optional[str] = None, return_indexes: Optional[bool] = False
    ):
        """Returns a slice of the tensor with only those elements that were modified/added.
        By default the modifications are calculated relative to the previous commit made, but this can be changed by providing a ``target id``.

        Args:
            target_id (str, optional): The commit id or branch name to calculate the modifications relative to. Defaults to ``None``.
            return_indexes (bool, optional): If ``True``, returns the indexes of the modified elements. Defaults to ``False``.

        Returns:
            Tensor: A new tensor with only the modified elements if ``return_indexes`` is ``False``.
            Tuple[Tensor, List[int]]: A new tensor with only the modified elements and the indexes of the modified elements if ``return_indexes`` is ``True``.

        Raises:
            TensorModifiedError: If a target id is passed which is not an ancestor of the current commit.
        """
        current_commit_id = self.version_state["commit_id"]
        indexes = get_modified_indexes(
            self.key,
            current_commit_id,
            target_id,
            self.version_state,
            self.storage,
        )
        tensor = self[indexes]
        if return_indexes:
            return tensor, indexes
        return tensor

    @property
    def meta(self):
        """Metadata of the tensor."""
        return self.chunk_engine.tensor_meta

    @property
    def shape(self) -> Tuple[Optional[int], ...]:
        """Get the shape of this tensor. Length is included.

        Example:

            >>> tensor.append(np.zeros((10, 10)))
            >>> tensor.append(np.zeros((10, 15)))
            >>> tensor.shape
            (2, 10, None)

        Returns:
            tuple: Tuple where each value is either ``None`` (if that axis is dynamic) or an int (if that axis is fixed).

        Note:
            If you don't want ``None`` in the output shape or want the lower/upper bound shapes,
            use :attr:`shape_interval` instead.
        """
        sample_shape_tensor = self._sample_shape_tensor
        sample_shape_provider = (
            self._sample_shape_provider(sample_shape_tensor)
            if sample_shape_tensor
            else None
        )
        shape: Tuple[Optional[int], ...]
        shape = self.chunk_engine.shape(
            self.index, sample_shape_provider=sample_shape_provider
        )
        if not shape and self.meta.max_shape:
            shape = (0,) * len(self.meta.max_shape)
        if self.meta.max_shape == [0, 0, 0]:
            shape = ()
        return shape

    @property
    def size(self) -> Optional[int]:
        s = 1
        for x in self.shape:
            if x is None:
                return None
            s *= x  # not using np.prod to avoid overflow
        return s

    @property
    def ndim(self) -> int:
        """Number of dimensions of the tensor."""
        return self.chunk_engine.ndim(self.index)

    @property
    def dtype(self) -> Optional[np.dtype]:
        """Dtype of the tensor."""
        if self.base_htype in ("json", "list"):
            return np.dtype(str)
        if self.meta.dtype:
            return np.dtype(self.meta.dtype)
        return None

    @property
    def is_sequence(self):
        """Whether this tensor is a sequence tensor."""
        return self.meta.is_sequence

    @property
    def is_link(self):
        """Whether this tensor is a link tensor."""
        return self.meta.is_link

    @property
    def verify(self):
        """Whether linked data will be verified when samples are added. Applicable only to tensors with htype ``link[htype]``."""
        return self.is_link and self.meta.verify

    @property
    def htype(self):
        """Htype of the tensor."""
        htype = self.meta.htype
        if self.is_sequence:
            htype = f"sequence[{htype}]"
        if self.is_link:
            htype = f"link[{htype}]"
        return htype

    @property
    def hidden(self) -> bool:
        """Whether this tensor is a hidden tensor."""
        return self.meta.hidden

    @property
    def base_htype(self):
        """Base htype of the tensor.

        Example:

            >>> ds.create_tensor("video_seq", htype="sequence[video]", sample_compression="mp4")
            >>> ds.video_seq.htype
            sequence[video]
            >>> ds.video_seq.base_htype
            video
        """
        return self.meta.htype

    @property
    def shape_interval(self) -> ShapeInterval:
        """Returns a :class:`~hub.util.shape_interval.ShapeInterval` object that describes this tensor's shape more accurately. Length is included.

        Example:

            >>> tensor.append(np.zeros((10, 10)))
            >>> tensor.append(np.zeros((10, 15)))
            >>> tensor.shape_interval
            ShapeInterval(lower=(2, 10, 10), upper=(2, 10, 15))
            >>> str(tensor.shape_interval)
            (2, 10, 10:15)

        Returns:
            ShapeInterval: Object containing ``lower`` and ``upper`` properties.

        Note:
            If you are expecting a tuple, use :attr:`shape` instead.
        """
        return self.chunk_engine.shape_interval

    @property
    def is_dynamic(self) -> bool:
        """Will return ``True`` if samples in this tensor have shapes that are unequal."""
        return self.shape_interval.is_dynamic

    @property
    def num_samples(self) -> int:
        """Returns the length of the primary axis of the tensor.
        Ignores any applied indexing and returns the total length.
        """
        if self.is_sequence:
            return self.chunk_engine._sequence_length
        return self.chunk_engine.num_samples

    def __len__(self):
        """Returns the length of the primary axis of the tensor.
        Accounts for indexing into the tensor object.

        Examples:
            >>> len(tensor)
            0
            >>> tensor.extend(np.zeros((100, 10, 10)))
            >>> len(tensor)
            100
            >>> len(tensor[5:10])
            5

        Returns:
            int: The current length of this tensor.
        """

        # catch corrupted datasets / user tampering ASAP
        self.chunk_engine.validate_num_samples_is_synchronized()

        return self.index.length(self.num_samples)

    def __getitem__(
        self,
        item: Union[int, slice, List[int], Tuple[Union[int, slice, Tuple[int]]], Index],
        is_iteration: bool = False,
    ):
        if not isinstance(item, (int, slice, list, tuple, Index)):
            raise InvalidKeyTypeError(item)
        return Tensor(
            self.key,
            self.dataset,
            index=self.index[item],
            is_iteration=is_iteration,
            chunk_engine=self.chunk_engine,
        )

    def _get_bigger_dtype(self, d1, d2):
        if np.can_cast(d1, d2):
            if np.can_cast(d2, d1):
                return d1
            else:
                return d2
        else:
            if np.can_cast(d2, d1):
                return d2
            else:
                return np.object

    def _infer_np_dtype(self, val: Any) -> np.dtype:
        # TODO refac
        if hasattr(val, "dtype"):
            return val.dtype
        elif isinstance(val, int):
            return np.array(0).dtype
        elif isinstance(val, float):
            return np.array(0.0).dtype
        elif isinstance(val, str):
            return np.array("").dtype
        elif isinstance(val, bool):
            return np.dtype(bool)
        elif isinstance(val, Sequence):
            return reduce(self._get_bigger_dtype, map(self._infer_np_dtype, val))
        else:
            raise TypeError(f"Cannot infer numpy dtype for {val}")

    def __setitem__(self, item: Union[int, slice], value: Any):
        """Update samples with new values.

        Example:

            >>> tensor.append(np.zeros((10, 10)))
            >>> tensor.shape
            (1, 10, 10)
            >>> tensor[0] = np.zeros((3, 3))
            >>> tensor.shape
            (1, 3, 3)
        """
        self._write_initialization()
        [f() for f in list(self.dataset._update_hooks.values())]
        update_link_callback = self._update_links if self.meta.links else None
        if isinstance(value, Tensor):
            if value._skip_next_setitem:
                value._skip_next_setitem = False
                return
            value = value.numpy(aslist=True)
        item_index = Index(item)

        if (
            hub.constants._ENABLE_RANDOM_ASSIGNMENT
            and isinstance(item, int)
            and item >= self.num_samples
        ):
            if self.is_sequence:
                raise NotImplementedError(
                    "Random assignment is not supported for sequences yet."
                )
            num_samples_to_pad = item - self.num_samples
            append_link_callback = self._append_to_links if self.meta.links else None

            self.chunk_engine.pad_and_append(
                num_samples_to_pad,
                value,
                append_link_callback=append_link_callback,
                update_link_callback=update_link_callback,
            )
        else:

            if not item_index.values[0].subscriptable() and not self.is_sequence:
                # we're modifying a single sample, convert it to a list as chunk engine expects multiple samples
                value = [value]

            self.chunk_engine.update(
                self.index[item_index],
                value,
                link_callback=update_link_callback,
            )
        dataset_written(self.dataset)

    def __iter__(self):
        for i in range(len(self)):
            yield self.__getitem__(i, is_iteration=True)

    def numpy(
        self, aslist=False, fetch_chunks=False
    ) -> Union[np.ndarray, List[np.ndarray]]:
        """Computes the contents of the tensor in numpy format.

        Args:
            aslist (bool): If ``True``, a list of np.ndarrays will be returned. Helpful for dynamic tensors.
                If ``False``, a single np.ndarray will be returned unless the samples are dynamically shaped, in which case
                an error is raised.
            fetch_chunks (bool): If ``True``, full chunks will be retrieved from the storage, otherwise only required bytes will be retrieved.
                This will always be ``True`` even if specified as ``False`` in the following cases:

                - The tensor is ChunkCompressed.
                - The chunk which is being accessed has more than 128 samples.

        Raises:
            DynamicTensorNumpyError: If reading a dynamically-shaped array slice without ``aslist=True``.
            ValueError: If the tensor is a link and the credentials are not populated.

        Returns:
            A numpy array containing the data represented by this tensor.
        """
        ret = self.chunk_engine.numpy(
            self.index,
            aslist=aslist,
            fetch_chunks=fetch_chunks,
            pad_tensor=self.pad_tensor,
        )
        if self.htype == "point_cloud":
            if isinstance(ret, list):
                ret = [arr[..., :3] for arr in ret]
            else:
                ret = ret[..., :3]
        dataset_read(self.dataset)
        return ret

    def summary(self):
        """Prints a summary of the tensor."""
        pretty_print = summary_tensor(self)

        print(self)
        print(pretty_print)

    def __str__(self):
        index_str = f", index={self.index}"
        if self.index.is_trivial():
            index_str = ""
        return f"Tensor(key={repr(self.meta.name or self.key)}{index_str})"

    __repr__ = __str__

    def __array__(self) -> np.ndarray:
        return self.numpy()  # type: ignore

    @_inplace_op
    def __iadd__(self, other):
        pass

    @_inplace_op
    def __isub__(self, other):
        pass

    @_inplace_op
    def __imul__(self, other):
        pass

    @_inplace_op
    def __itruediv__(self, other):
        pass

    @_inplace_op
    def __ifloordiv__(self, other):
        pass

    @_inplace_op
    def __imod__(self, other):
        pass

    @_inplace_op
    def __ipow__(self, other):
        pass

    @_inplace_op
    def __ilshift__(self, other):
        pass

    @_inplace_op
    def __irshift__(self, other):
        pass

    @_inplace_op
    def __iand__(self, other):
        pass

    @_inplace_op
    def __ixor__(self, other):
        pass

    @_inplace_op
    def __ior__(self, other):
        pass

    def data(self, aslist: bool = False) -> Any:
        """Returns data in the tensor in a format based on the tensor's base htype.

        - Returns dict with dict["value"] = :meth:`Tensor.text() <text>` for tensors with base htype of 'text'.

        - Returns dict with dict["value"] = :meth:`Tensor.dict() <dict>` for tensors with base htype of 'json'.

        - Returns dict with dict["value"] = :meth:`Tensor.list() <list>` for tensors with base htype of 'list'.

        - For video tensors, returns a dict with keys "frames", "timestamps" and "sample_info":

            - Value of dict["frames"] will be same as :meth:`numpy`.
            - Value of dict["timestamps"] will be same as :attr:`timestamps` corresponding to the frames.
            - Value of dict["sample_info"] will be same as :attr:`sample_info`.

        - For class_label tensors, returns a dict with keys "value" and "text".

            - Value of dict["value"] will be same as :meth:`numpy`.
            - Value of dict["text"] will be list of class labels as strings.

        - For image or dicom tensors, returns dict with keys "value" and "sample_info".

            - Value of dict["value"] will be same as :meth:`numpy`.
            - Value of dict["sample_info"] will be same as :attr:`sample_info`.

        - For all else, returns dict with key "value" with value same as :meth:`numpy`.
        """
        htype = self.base_htype
        if htype == "text":
            return {"value": self.text()}
        if htype == "json":
            return {"value": self.dict()}
        if htype == "list":
            return {"value": self.list()}
        if self.htype == "video":
            data = {}
            data["frames"] = self.numpy(aslist=aslist)
            index = self.index
            if index.values[0].subscriptable():
                root = Tensor(self.key, self.dataset)
                if len(index.values) > 1:
                    data["timestamps"] = np.array(
                        [
                            root[i, index.values[1].value].timestamps  # type: ignore
                            for i in index.values[0].indices(self.num_samples)
                        ]
                    )
                else:
                    data["timestamps"] = np.array(
                        [
                            root[i].timestamps
                            for i in index.values[0].indices(self.num_samples)
                        ]
                    )
            else:
                data["timestamps"] = self.timestamps
            if aslist:
                data["timestamps"] = data["timestamps"].tolist()  # type: ignore

            data["sample_info"] = self.sample_info  # type: ignore
            return data
        if htype == "class_label":
            labels = self.numpy(aslist=aslist)
            data = {"value": labels}
            class_names = self.info.class_names
            if class_names:
                data["text"] = convert_to_text(labels, self.info.class_names)
            return data
        if htype in ("image", "image.rgb", "image.gray", "dicom"):
            return {
                "value": self.numpy(aslist=aslist),
                "sample_info": self.sample_info or {},
            }
        elif htype == "point_cloud":
            full_arr = self.chunk_engine.numpy(
                self.index,
                aslist=aslist,
                pad_tensor=self.pad_tensor,
            )

            if self.ndim == 2:
                meta = {}  # type: ignore

                if len(self.sample_info) == 0:
                    return meta

                for i, dimension_name in enumerate(self.sample_info["dimension_names"]):  # type: ignore
                    typestr = POINT_CLOUD_FIELD_NAME_TO_TYPESTR[dimension_name]
                    meta[dimension_name] = full_arr[..., i].astype(np.dtype(typestr))  # type: ignore
                return meta

            meta = []  # type: ignore
            for sample_index in range(len(full_arr)):
                meta_dict = {}  # type: ignore

                if len(self.sample_info[sample_index]) == 0:
                    meta.append(meta_dict)  # type: ignore
                    continue

                for dimension_index, dimension_name in enumerate(
                    self.sample_info[sample_index]["dimension_names"]
                ):
                    dtype = POINT_CLOUD_FIELD_NAME_TO_TYPESTR[dimension_name]
                    meta_dict[dimension_name] = cast_point_cloud_array_to_proper_dtype(
                        full_arr, sample_index, dimension_index, dtype
                    )
                meta.append(meta_dict)  # type: ignore

            if len(full_arr) == 1:
                meta = meta[0]  # type: ignore
            return meta

        else:
            return {
                "value": self.numpy(aslist=aslist),
            }

    def tobytes(self) -> bytes:
        """Returns the bytes of the tensor.

        - Only works for a single sample of tensor.
        - If the tensor is uncompressed, this returns the bytes of the numpy array.
        - If the tensor is sample compressed, this returns the compressed bytes of the sample.
        - If the tensor is chunk compressed, this raises an error.

        Returns:
            bytes: The bytes of the tensor.

        Raises:
            ValueError: If the tensor has multiple samples.
        """
        if self.index.values[0].subscriptable() or len(self.index.values) > 1:
            raise ValueError("tobytes() can be used only on exatcly 1 sample.")
        idx = self.index.values[0].value
        ret = self.chunk_engine.read_bytes_for_sample(idx)  # type: ignore
        dataset_read(self.dataset)
        return ret

    def _append_to_links(self, sample, flat: Optional[bool]):
        for k, v in self.meta.links.items():
            if flat is None or v["flatten_sequence"] == flat:
                v = get_link_transform(v["append"])(sample, self.link_creds)
                tensor = Tensor(k, self.dataset)
                if (
                    isinstance(v, np.ndarray)
                    and tensor.dtype
                    and v.dtype != tensor.dtype
                ):
                    v = v.astype(tensor.dtype)  # bc
                tensor.append(v)

    def _update_links(
        self,
        global_sample_index: int,
        sub_index: Index,
        new_sample,
        flat: Optional[bool],
    ):
        for k, v in self.meta.links.items():
            if flat is None or v["flatten_sequence"] == flat:
                fname = v.get("update")
                if fname:
                    func = get_link_transform(fname)
                    tensor = Tensor(k, self.dataset)
                    val = func(
                        new_sample,
                        tensor[global_sample_index],
                        sub_index=sub_index,
                        partial=not sub_index.is_trivial(),
                        link_creds=self.link_creds,
                    )
                    if val is not _NO_LINK_UPDATE:
                        if (
                            isinstance(val, np.ndarray)
                            and tensor.dtype
                            and val.dtype != tensor.dtype
                        ):
                            val = val.astype(tensor.dtype)  # bc
                        tensor[global_sample_index] = val

    @property
    def _sample_info_tensor(self):
        ds = self.dataset
        return ds.version_state["full_tensors"].get(
            ds.version_state["tensor_names"].get(get_sample_info_tensor_key(self.key))
        )

    @property
    def _sample_shape_tensor(self):
        ds = self.dataset
        return ds.version_state["full_tensors"].get(
            ds.version_state["tensor_names"].get(get_sample_shape_tensor_key(self.key))
        )

    @property
    def _sample_id_tensor(self):
        return self.dataset._tensors().get(get_sample_id_tensor_key(self.key))

    def _sample_shape_provider(self, sample_shape_tensor) -> Callable:
        if self.is_sequence:

            def get_sample_shape(global_sample_index: int):
                seq_pos = slice(
                    *self.chunk_engine.sequence_encoder[global_sample_index]
                )
                idx = Index([IndexEntry(seq_pos)])
                shapes = sample_shape_tensor[idx].numpy()
                return shapes

        else:

            def get_sample_shape(global_sample_index: int):
                return tuple(sample_shape_tensor[global_sample_index].numpy().tolist())

        return get_sample_shape

    def _get_sample_info_at_index(self, global_sample_index: int, sample_info_tensor):
        if self.is_sequence:
            return [
                sample_info_tensor[i].data()
                for i in range(*self.chunk_engine.sequence_encoder[global_sample_index])
            ]
        return sample_info_tensor[global_sample_index].data()["value"]

    def _sample_info(self, index: Index):
        sample_info_tensor = self._sample_info_tensor
        if sample_info_tensor is None:
            return None
        if index.subscriptable_at(0):
            return list(
                map(
                    partial(
                        self._get_sample_info_at_index,
                        sample_info_tensor=sample_info_tensor,
                    ),
                    index.values[0].indices(self.num_samples),
                )
            )
        return self._get_sample_info_at_index(index.values[0].value, sample_info_tensor)  # type: ignore

    @property
    def sample_info(self) -> Union[Dict, List[Dict]]:
        """Returns info about particular samples in a tensor. Returns dict in case of single sample, otherwise list of dicts.
        Data in returned dict would depend on the tensor's htype and the sample itself.

        Example:

            >>> ds.videos[0].sample_info
            {'duration': 400400, 'fps': 29.97002997002997, 'timebase': 3.3333333333333335e-05, 'shape': [400, 360, 640, 3], 'format': 'mp4', 'filename': '../hub/tests/dummy_data/video/samplemp4.mp4', 'modified': False}
            >>> ds.images[:2].sample_info
            [{'exif': {'Software': 'Google'}, 'shape': [900, 900, 3], 'format': 'jpeg', 'filename': '../hub/tests/dummy_data/images/cat.jpeg', 'modified': False}, {'exif': {}, 'shape': [495, 750, 3], 'format': 'jpeg', 'filename': '../hub/tests/dummy_data/images/car.jpg', 'modified': False}]
        """
        return self._sample_info(self.index)

    def _linked_sample(self):
        if not self.is_link:
            raise ValueError("Not supported as the tensor is not a link.")
        if self.index.values[0].subscriptable() or len(self.index.values) > 1:
            raise ValueError("_linked_sample can be used only on exatcly 1 sample.")
        return self.chunk_engine.linked_sample(self.index.values[0].value)

    def _get_video_stream_url(self):
        if self.is_link:
            return self.chunk_engine.get_video_url(self.index.values[0].value)

        from hub.visualizer.video_streaming import get_video_stream_url

        return get_video_stream_url(self, self.index.values[0].value)

    def play(self):
        """Play video sample. Plays video in Jupyter notebook or plays in web browser. Video is streamed directly from storage.
        This method will fail for incompatible htypes.

        Example:

            >>> ds = hub.load("./test/my_video_ds")
            >>> # play second sample
            >>> ds.videos[2].play()

        Note:
            Video streaming is not yet supported on colab.
        """
        if (
            get_compression_type(self.meta.sample_compression) != VIDEO_COMPRESSION
            and self.htype != "link[video]"
        ):
            raise Exception("Only supported for video tensors.")
        if self.index.values[0].subscriptable():
            raise ValueError("Video streaming requires exactly 1 sample.")
        if len(self.index.values) > 1:
            warnings.warn(
                "Sub indexes to video sample will be ignored while streaming."
            )
        if is_colab():
            raise NotImplementedError("Video streaming is not supported on colab yet.")
        elif is_jupyter():
            return video_html(
                src=self._get_video_stream_url(),
                alt=f"{self.key}[{self.index.values[0].value}]",
            )
        else:
            webbrowser.open(self._get_video_stream_url())

    @invalid_view_op
    def pop(self, index: Optional[int] = None):
        """Removes an element at the given index."""
        if index is None:
            index = self.num_samples - 1
        self.chunk_engine.pop(index)
        [self.dataset[link].pop(index) for link in self.meta.links]

    @property
    def timestamps(self) -> np.ndarray:
        """Returns timestamps (in seconds) for video sample as numpy array.

        Example:

            >>> # Return timestamps for all frames of first video sample
            >>> ds.videos[0].timestamps.shape
            (400,)
            >>> # Return timestamps for 5th to 10th frame of first video sample
            >>> ds.videos[0, 5:10].timestamps
            array([0.2002    , 0.23356667, 0.26693332, 0.33366665, 0.4004    ],
            dtype=float32)

        """
        if (
            get_compression_type(self.meta.sample_compression) != VIDEO_COMPRESSION
            and self.htype != "link[video]"
        ):
            raise Exception("Only supported for video tensors.")
        index = self.index
        if index.values[0].subscriptable():
            raise ValueError("Only supported for exactly 1 video sample.")
        if self.is_sequence:
            if len(index.values) == 1 or index.values[1].subscriptable():
                raise ValueError("Only supported for exactly 1 video sample.")
            sub_index = index.values[2].value if len(index.values) > 2 else None
        else:
            sub_index = index.values[1].value if len(index.values) > 1 else None
        global_sample_index = next(index.values[0].indices(self.num_samples))
        if self.is_link:
            sample = self.chunk_engine.get_video_url(global_sample_index)  # type: ignore
        else:
            sample = self.chunk_engine.get_video_sample(
                global_sample_index, index, decompress=False
            )

        nframes = self.shape[0]
        start, stop, step, reverse = normalize_index(sub_index, nframes)

        stamps = _read_timestamps(sample, start, stop, step, reverse)
        return stamps

    @property
    def _config(self):
        """Returns a summary of the configuration of the tensor."""
        tensor_meta = self.meta
        return {
            "htype": tensor_meta.htype or UNSPECIFIED,
            "dtype": tensor_meta.dtype or UNSPECIFIED,
            "sample_compression": tensor_meta.sample_compression or UNSPECIFIED,
            "chunk_compression": tensor_meta.chunk_compression or UNSPECIFIED,
            "hidden": tensor_meta.hidden,
            "is_link": tensor_meta.is_link,
            "is_sequence": tensor_meta.is_sequence,
        }

    @property
    def sample_indices(self):
        """Returns all the indices pointed to by this tensor in the dataset view."""
        return self.dataset._sample_indices(self.num_samples)

    def _extract_value(self, htype):
        if self.base_htype != htype:
            raise Exception(f"Only supported for {htype} tensors.")

        if self.ndim == 1:
            return self.numpy()[0]
        else:
            return [sample[0] for sample in self.numpy(aslist=True)]

    def text(self):
        """Return text data. Only applicable for tensors with 'text' base htype."""
        return self._extract_value("text")

    def dict(self):
        """Return json data. Only applicable for tensors with 'json' base htype."""
        return self._extract_value("json")

    def list(self):
        """Return list data. Only applicable for tensors with 'list' base htype."""
        if self.base_htype != "list":
            raise Exception(f"Only supported for list tensors.")

        if self.ndim == 1:
            return list(self.numpy())
        else:
            return list(map(list, self.numpy(aslist=True)))



import pytest

import hub
from hub.core.dataset import Dataset
from hub.core.compression import compress_multiple
from hub.tests.common import get_dummy_data_path
from hub.util.exceptions import CorruptedSampleError

import numpy as np


def test_point_cloud(local_ds, point_cloud_paths):
    for i, (compression, path) in enumerate(point_cloud_paths.items()):
        if compression == "las":
            tensor = local_ds.create_tensor(
                f"point_cloud_{i}", htype="point_cloud", sample_compression=compression
            )
            sample = hub.read(path)

            if "point_cloud" in path:  # check shape only for internal test point_clouds
                assert sample.shape[0] == 20153

            assert len(sample.meta) == 6
            assert len(sample.meta["dimension_names"]) == 18
            assert len(sample.meta["las_header"]) == 19

            assert sample.meta["las_header"]["DEFAULT_VERSION"] == {
                "major": 1,
                "minor": 2,
            }
            assert sample.meta["las_header"]["creation_date"] == {
                "year": 2022,
                "month": 5,
                "day": 24,
            }
            assert sample.meta["las_header"]["version"] == {"major": 1, "minor": 2}
            assert (
                sample.meta["las_header"]["uuid"]
                == "00000000-0000-0000-0000-000000000000"
            )

            tensor.append(sample)
            tensor.append(sample)
            tensor.append(sample)
            assert tensor.shape == (3, 20153, 18)

            shape_tester(local_ds, path, sample, tensor, feature_size=18)

            with pytest.raises(NotImplementedError):
                arrays = np.zeros((5, 1000, 3))
                compress_multiple(arrays, compression)

    local_ds.create_tensor(
        "point_cloud_without_sample_compression",
        htype="point_cloud",
        sample_compression=None,
    )
    local_ds.point_cloud_without_sample_compression.append(
        np.zeros((1000, 3), dtype=np.float32)
    )
    np.testing.assert_array_equal(
        local_ds.point_cloud_without_sample_compression[0].numpy(),
        np.zeros((1000, 3), dtype=np.float32),
    )
    local_ds.point_cloud_without_sample_compression.data()
    assert len(local_ds.point_cloud_without_sample_compression.data()) == 0

    local_ds.point_cloud_without_sample_compression.append(hub.read(path))
    assert local_ds.point_cloud_without_sample_compression[1].numpy().shape == (
        20153,
        3,
    )
    assert len(local_ds.point_cloud_without_sample_compression.numpy(aslist=True)) == 2
    assert len(local_ds.point_cloud_without_sample_compression.data(aslist=True)) == 2
    local_ds.create_tensor(
        "point_cloud_with_sample_compression",
        htype="point_cloud",
        sample_compression="las",
    )
    with pytest.raises(NotImplementedError):
        local_ds.point_cloud_with_sample_compression.append(
            np.zeros((1000, 3), dtype=np.float32)
        )

    with pytest.raises(CorruptedSampleError):
        local_ds.point_cloud_with_sample_compression.append(
            hub.read(get_dummy_data_path("point_cloud/corrupted_point_cloud.las"))
        )

    local_ds.point_cloud_with_sample_compression.append(hub.read(path, verify=True))
    assert local_ds.point_cloud_with_sample_compression.shape == (1, 20153, 18)

    local_ds.create_tensor(
        "point_cloud_data_method_type_tester",
        htype="point_cloud",
        sample_compression="las",
    )
    local_ds.point_cloud_data_method_type_tester.append(sample)
    assert isinstance(local_ds.point_cloud_data_method_type_tester.data(), dict)

    local_ds.point_cloud_data_method_type_tester.append(sample)
    assert isinstance(
        local_ds.point_cloud_data_method_type_tester.data(aslist=True), list
    )


def shape_tester(local_ds, path, sample, tensor, feature_size):
    with local_ds:
        for _ in range(5):
            tensor.append(hub.read(path))  # type: ignore
        tensor.extend([hub.read(path) for _ in range(5)])  # type: ignore

    for i in range(10):
        assert tensor[i].numpy().shape[0] == sample.shape[0]  # type: ignore
        assert len(tensor[i].data()) == feature_size


def test_point_cloud_slicing(local_ds: Dataset, point_cloud_paths):
    for compression, path in point_cloud_paths.items():
        if compression == "las":
            dummy = np.zeros((20153, 3))
        local_ds.create_tensor(
            "point_cloud", htype="point_cloud", sample_compression=compression
        )
        local_ds.point_cloud.append(hub.read(path))
        assert local_ds.point_cloud[0][0:5].numpy().shape == dummy[0:5].shape
        assert local_ds.point_cloud[0][100:120].numpy().shape == dummy[100:120].shape
        assert local_ds.point_cloud[0][120].numpy().shape == dummy[120].shape
        assert local_ds.point_cloud[0][-1].numpy().shape == dummy[-1].shape
        assert local_ds.point_cloud[0][10:5:-2].numpy().shape == dummy[10:5:-2].shape
        assert (
            local_ds.point_cloud[0][-3:-10:-1].numpy().shape == dummy[-3:-10:-1].shape
        )
        assert (
            local_ds.point_cloud[0][-25:100:-2].numpy().shape == dummy[-25:100:-2].shape
        )
        assert local_ds.point_cloud[0][::-1].numpy().shape == dummy[::-1].shape
        assert local_ds.point_cloud[0][:5:-1].numpy().shape == dummy[:5:-1].shape
        assert local_ds.point_cloud[0][-1].numpy().shape == dummy[-1].shape
        return



import json
import os
import warnings
import requests
from hub.client.config import HUB_PYPI_VERSION_PATH
from hub.core.fast_forwarding import version_compare
import time


def get_latest_version():
    if os.path.exists(HUB_PYPI_VERSION_PATH):
        with open(HUB_PYPI_VERSION_PATH) as f:
            latest_version, saved_time = json.load(f)
            seconds_in_a_day = 60 * 60 * 24
            time_elapsed = time.time() - saved_time
            if time_elapsed < seconds_in_a_day:
                return latest_version

    response = requests.get("https://pypi.org/pypi/hub/json")
    latest_version = response.json()["info"]["version"]
    with open(HUB_PYPI_VERSION_PATH, "w") as f:
        json.dump((latest_version, time.time()), f)
    return latest_version


def warn_if_update_required(current_version):
    try:
        latest_version = get_latest_version()
    except Exception:
        return
    if version_compare(current_version, latest_version) < 0:
        warnings.warn(
            f"A newer version of hub ({latest_version}) is available. It's recommended that you update to the latest version using `pip install -U hub`."
        )



from typing import Optional, Sequence, List
from hub.util.exceptions import ReadOnlyModeError, TensorDoesNotExistError


def try_flushing(ds):
    try:
        ds.flush()
    except ReadOnlyModeError:
        pass


def map_tensor_keys(dataset, tensor_keys: Optional[Sequence[str]] = None) -> List[str]:
    """Sanitizes tensor_keys if not None, else returns all the keys present in the dataset."""

    tensors = dataset.tensors

    if tensor_keys is None:
        tensor_keys = list(tensors)
    else:
        for t in tensor_keys:
            if t not in tensors:
                raise TensorDoesNotExistError(t)

        tensor_keys = list(tensor_keys)

    # Get full path in case of groups
    return [tensors[k].meta.name or tensors[k].key for k in tensor_keys]



from abc import ABC, abstractmethod


class ComputeProvider(ABC):
    """An abstract base class for implementing a compute provider."""

    def __init__(self, workers):
        self.workers = workers

    def map_with_progressbar(self, func, iterable, total_length: int, desc: str = None):
        from tqdm.std import tqdm  # type: ignore
        import threading
        from threading import Thread

        progress_bar = tqdm(total=total_length, desc=desc)
        progress_queue = self.create_queue()

        def sub_func(*args, **kwargs):
            def pg_callback(value: int):
                progress_queue.put(value)  # type: ignore[trust]
                pass

            return func(pg_callback, *args, **kwargs)

        def update_pg(bar, queue):
            while True:
                r = queue.get()
                if r is not None:
                    bar.update(r)
                else:
                    break

        progress_thread: Thread = threading.Thread(
            target=update_pg, args=(progress_bar, progress_queue)
        )
        progress_thread.start()

        try:
            result = self.map(sub_func, iterable)
        finally:
            progress_queue.put(None)  # type: ignore[trust]
            progress_thread.join()
            progress_bar.close()

            if hasattr(progress_queue, "close"):
                progress_bar.close()

        return result

    @abstractmethod
    def create_queue(self):
        """Creates queue for specific provider"""

    @abstractmethod
    def map(self, func, iterable):
        """Applies 'func' to each element in 'iterable', collecting the results
        in a list that is returned.
        """

    @abstractmethod
    def close(self):
        """Closes the provider."""



from .dataset import Dataset  # type: ignore
from .hub_cloud_dataset import HubCloudDataset
from .view_entry import ViewEntry

from hub.util.path import is_hub_cloud_path

# NOTE: experimentation helper
FORCE_CLASS = None


def dataset_factory(path, *args, **kwargs):
    """Returns a Dataset object from the appropriate class. For example: If `path` is a hub
    cloud path (prefixed with `hub://`), the returned Dataset object will be of HubCloudDataset.
    """
    if FORCE_CLASS is not None:
        clz = FORCE_CLASS
    elif is_hub_cloud_path(path):
        clz = HubCloudDataset
    else:
        clz = Dataset

    if clz in {Dataset, HubCloudDataset}:
        ds = clz(path=path, *args, **kwargs)
        if ds.info.get("virtual-datasource", False):
            ds = ds._get_view()
        return ds
    raise TypeError(f"Invalid dataset class {clz}")



import hub
from hub.core.version_control.commit_node import CommitNode
from hub.constants import FIRST_COMMIT_ID
from hub.util.version_control import (
    _merge_commit_node_maps,
    _version_info_to_json,
    _version_info_from_json,
)


def test_merge_commit_node_map():

    root = CommitNode("main", FIRST_COMMIT_ID)
    a = CommitNode("main", "a")
    b = CommitNode("main", "b")
    c = CommitNode("main", "c")
    e = CommitNode("main", "e")
    root.add_successor(a, "commit a")
    root.add_successor(b, "commit b")
    a.add_successor(c, "commit c")
    c.add_successor(e, "commit e")
    map1 = {
        FIRST_COMMIT_ID: root,
        "a": a,
        "b": b,
        "c": c,
        "e": e,
    }

    root = CommitNode("main", FIRST_COMMIT_ID)
    a = CommitNode("main", "a")
    b = CommitNode("main", "b")
    d = CommitNode("main", "d")
    f = CommitNode("main", "f")
    root.add_successor(a, "commit a")
    root.add_successor(b, "commit b")
    b.add_successor(d, "commit d")
    d.add_successor(f, "commit f")

    map2 = {
        FIRST_COMMIT_ID: root,
        "a": a,
        "b": b,
        "d": d,
        "f": f,
    }

    merged = _merge_commit_node_maps(map1, map2)

    assert set(merged.keys()) == set((FIRST_COMMIT_ID, "a", "b", "c", "d", "e", "f"))
    get_children = lambda node: set(c.commit_id for c in node.children)
    assert get_children(merged[FIRST_COMMIT_ID]) == set(("a", "b"))
    assert get_children(merged["a"]) == set(("c"))
    assert get_children(merged["b"]) == set(("d"))
    assert get_children(merged["c"]) == set(("e"))
    assert get_children(merged["d"]) == set(("f"))

    # Test json encoding
    branch_commit_map = {"main": "f"}
    version_info = {"commit_node_map": merged, "branch_commit_map": branch_commit_map}
    encoded = _version_info_to_json(version_info)
    decoded = _version_info_from_json(encoded)
    assert decoded["branch_commit_map"] == branch_commit_map
    merged = decoded["commit_node_map"]

    assert set(merged.keys()) == set((FIRST_COMMIT_ID, "a", "b", "c", "d", "e", "f"))
    get_children = lambda node: set(c.commit_id for c in node.children)
    assert get_children(merged[FIRST_COMMIT_ID]) == set(("a", "b"))
    assert get_children(merged["a"]) == set(("c"))
    assert get_children(merged["b"]) == set(("d"))
    assert get_children(merged["c"]) == set(("e"))
    assert get_children(merged["d"]) == set(("f"))



from abc import ABC, abstractmethod
from collections.abc import MutableMapping
from typing import Optional, Set, Sequence

from hub.constants import BYTE_PADDING
from hub.util.assert_byte_indexes import assert_byte_indexes
from hub.util.exceptions import ReadOnlyModeError
from hub.util.keys import get_dataset_lock_key
import posixpath


class StorageProvider(ABC, MutableMapping):
    autoflush = False
    read_only = False
    root = ""
    _is_hub_path = False

    """An abstract base class for implementing a storage provider.

    To add a new provider using Provider, create a subclass and implement all 5 abstract methods below.
    """

    @abstractmethod
    def __getitem__(self, path: str):
        """Gets the object present at the path within the given byte range.

        Args:
            path (str): The path relative to the root of the provider.

        Returns:
            bytes: The bytes of the object present at the path.

        Raises:
            KeyError: If an object is not found at the path.
        """

    def get_bytes(
        self,
        path: str,
        start_byte: Optional[int] = None,
        end_byte: Optional[int] = None,
    ):
        """Gets the object present at the path within the given byte range.

        Args:
            path (str): The path relative to the root of the provider.
            start_byte (int, optional): If only specific bytes starting from start_byte are required.
            end_byte (int, optional): If only specific bytes up to end_byte are required.

        Returns:
            bytes: The bytes of the object present at the path within the given byte range.

        Raises:
            InvalidBytesRequestedError: If `start_byte` > `end_byte` or `start_byte` < 0 or `end_byte` < 0.
            KeyError: If an object is not found at the path.
        """
        assert_byte_indexes(start_byte, end_byte)
        return self[path][start_byte:end_byte]

    @abstractmethod
    def __setitem__(self, path: str, value: bytes):
        """Sets the object present at the path with the value

        Args:
            path (str): the path relative to the root of the provider.
            value (bytes): the value to be assigned at the path.
        """

    def set_bytes(
        self,
        path: str,
        value: bytes,
        start_byte: Optional[int] = None,
        overwrite: Optional[bool] = False,
    ):
        """Sets the object present at the path with the value

        Args:
            path (str): the path relative to the root of the provider.
            value (bytes): the value to be assigned at the path.
            start_byte (int, optional): If only specific bytes starting from start_byte are to be assigned.
            overwrite (boolean, optional): If the value is True, if there is an object present at the path
                it is completely overwritten, without fetching it's data.

        Raises:
            InvalidBytesRequestedError: If `start_byte` < 0.
            ReadOnlyError: If the provider is in read-only mode.
        """
        self.check_readonly()
        start_byte = start_byte or 0
        end_byte = start_byte + len(value)
        assert_byte_indexes(start_byte, end_byte)

        if path in self and not overwrite:
            current_value = bytearray(self[path])
            # need to pad with zeros at the end to write extra bytes
            if end_byte > len(current_value):
                current_value = current_value.ljust(end_byte, BYTE_PADDING)
            current_value[start_byte:end_byte] = value
            self[path] = current_value
        else:
            # need to pad with zeros at the start to write from an offset
            if start_byte != 0:
                value = value.rjust(end_byte, BYTE_PADDING)
            self[path] = value

    @abstractmethod
    def __iter__(self):
        """Generator function that iterates over the keys of the provider.

        Yields:
            str: the path of the object that it is iterating over, relative to the root of the provider.
        """

    @abstractmethod
    def _all_keys(self) -> Set[str]:
        """Generator function that iterates over the keys of the provider.

        Returns:
            set: set of all keys present at the root of the provider.
        """

    @abstractmethod
    def __delitem__(self, path: str):
        """Delete the object present at the path.

        Args:
            path (str): the path to the object relative to the root of the provider.

        Raises:
            KeyError: If an object is not found at the path.
        """

    @abstractmethod
    def __len__(self):
        """Returns the number of files present inside the root of the provider.

        Returns:
            int: the number of files present inside the root.
        """

    def enable_readonly(self):
        """Enables read-only mode for the provider."""
        self.read_only = True

    def disable_readonly(self):
        """Disables read-only mode for the provider."""
        self.read_only = False

    def check_readonly(self):
        """Raises an exception if the provider is in read-only mode."""
        if self.read_only:
            raise ReadOnlyModeError()

    def flush(self):
        """Only needs to be implemented for caches. Flushes the data to the next storage provider.
        Should be a no op for Base Storage Providers like local, s3, azure, gcs, etc.
        """
        self.check_readonly()

    def maybe_flush(self):
        """Flush cache if autoflush has been enabled.
        Called at the end of methods which write data, to ensure consistency as a default.
        """
        if self.autoflush:
            self.flush()

    @abstractmethod
    def clear(self, prefix=""):
        """Delete the contents of the provider."""

    def delete_multiple(self, paths: Sequence[str]):
        for path in paths:
            del self[path]

    def empty(self) -> bool:
        lock_key = get_dataset_lock_key()
        return len(self) - int(lock_key in self) <= 0

    def copy(self):
        """Returns a copy of the provider.

        Returns:
            StorageProvider: A copy of the provider.
        """
        cls = self.__class__
        new_provider = cls.__new__(cls)
        new_provider.__setstate__(self.__getstate__())
        return new_provider

    def get_presigned_url(self, key: str) -> str:
        return posixpath.join(self.root, key)

    def get_object_size(self, key: str) -> int:
        raise NotImplementedError()
